!******************************************************************
!>
!>   Module to solve the 1D ADS equation using the Kurganov scheme
!>   combined with an explicit RK4 for time integration
!>   @author
!>   Arnau Folch
!>
!******************************************************************
     MODULE ADS
     use KindType
     use Domain
#if defined WITH_ACC
     use openacc
#endif
     implicit none
!
!>   type ADS_CB: control block for ADS solver
!
     type ADS_CB
     !
     procedure(bconditions), pointer, nopass :: apply_bconditions
     procedure(fluxlimiter), pointer, nopass :: flux_limiter
     !
     integer(ip) :: time_marching
     real(rp)    :: bvalue(2)
#if defined WITH_ACC
     integer(ip)     :: limitertype
     real(rp),pointer   :: iw(:,:,:)
#else
     real(rp), allocatable :: u(:,:,:)
     real(rp), allocatable :: v(:,:,:)
     real(rp), allocatable :: w(:,:,:)
     real(rp), allocatable :: raw_c (:)
     real(rp), allocatable :: new_c (:)
     real(rp), allocatable :: c0 (:)
     !
     ! Temporary arrays for KT_RHS
     !
     real(rp), allocatable :: raw_r (:)
     real(rp), allocatable :: raw_c_l (:)
     real(rp), allocatable :: raw_c_r (:)
     !
#endif
     end type ADS_CB
!
!    LIST OF PRIVATE VARIABLES
!
  type(ADS_CB), PRIVATE, TARGET :: CB    !> Variables related to ADS solver
  !
  integer(ip),  PRIVATE, parameter :: TIME_MARCHING_EULER = 1
  integer(ip),  PRIVATE, parameter :: TIME_MARCHING_RK    = 2
#if defined WITH_ACC
     real(rp), PRIVATE, allocatable :: xki(:)
     real(rp), PRIVATE, allocatable :: yki(:)
     real(rp), PRIVATE, allocatable :: zki(:)
     real(rp), PRIVATE, allocatable :: ri(:)
     real(rp), PRIVATE, allocatable :: rj(:)
     real(rp), PRIVATE, allocatable :: rk(:)
     real(rp), PRIVATE, allocatable :: Wswapi(:,:,:)
     real(rp), PRIVATE, allocatable :: Wswapo(:,:,:)
     real(rp), PRIVATE, allocatable :: Eswapi(:,:,:)
     real(rp), PRIVATE, allocatable :: Eswapo(:,:,:)
     real(rp), PRIVATE, allocatable :: Sswapi(:,:,:)
     real(rp), PRIVATE, allocatable :: Sswapo(:,:,:)
     real(rp), PRIVATE, allocatable :: Nswapi(:,:,:)
     real(rp), PRIVATE, allocatable :: Nswapo(:,:,:)
     real(rp), PRIVATE, allocatable :: Dswapi(:,:,:)
     real(rp), PRIVATE, allocatable :: Dswapo(:,:,:)
     real(rp), PRIVATE, allocatable :: Uswapi(:,:,:)
     real(rp), PRIVATE, allocatable :: Uswapo(:,:,:)
     real(rp), PRIVATE, allocatable :: dX1_p(:)
     real(rp), PRIVATE, allocatable :: dX1_b(:)
     real(rp), PRIVATE, allocatable :: dX2_p(:)
     real(rp), PRIVATE, allocatable :: dX2_b(:)
     real(rp), PRIVATE, allocatable :: dX3_p(:)
     real(rp), PRIVATE, allocatable :: dX3_b(:)
     real(rp), PRIVATE, allocatable :: s_kflux(:)
     real(rp), PRIVATE, allocatable :: e_kflux(:)
     real(rp), PRIVATE, allocatable :: s_iflux(:)
     real(rp), PRIVATE, allocatable :: e_iflux(:)
     real(rp), PRIVATE, TARGET, pinned, allocatable :: ADS_u(:,:,:)
     real(rp), PRIVATE, TARGET, pinned, allocatable :: ADS_v(:,:,:)
     real(rp), PRIVATE, TARGET, pinned, allocatable :: ADS_w(:,:,:,:)
     real(rp), PRIVATE, TARGET, pinned, allocatable :: ADS_k1(:,:,:)
     real(rp), PRIVATE, TARGET, pinned, allocatable :: ADS_k2(:,:,:)
     real(rp), PRIVATE, TARGET, pinned, allocatable :: ADS_k3(:,:,:)

     real(rp), PRIVATE, allocatable :: xnewc(:,:,:)
     real(rp), PRIVATE, allocatable :: xc0(:,:,:)
     real(rp), PRIVATE, allocatable :: ynewc(:,:,:)
     real(rp), PRIVATE, allocatable :: yc0(:,:,:)
     real(rp), PRIVATE, allocatable :: yinic(:,:,:)
     real(rp), PRIVATE, allocatable :: znewc(:,:,:)
     real(rp), PRIVATE, allocatable :: zc0(:,:,:)
     real(rp), PRIVATE, allocatable :: zinic(:,:,:)

#else
     real(rp), PRIVATE, TARGET, allocatable :: ADS_u(:,:,:)
     real(rp), PRIVATE, TARGET, allocatable :: ADS_v(:,:,:)
     real(rp), PRIVATE, TARGET, allocatable :: ADS_w(:,:,:,:)
     real(rp), PRIVATE, TARGET, allocatable :: ADS_k1(:,:,:)
     real(rp), PRIVATE, TARGET, allocatable :: ADS_k2(:,:,:)
     real(rp), PRIVATE, TARGET, allocatable :: ADS_k3(:,:,:)
#endif
!
!    LIST OF PUBLIC ROUTINES IN THE MODULE
!
     PUBLIC :: ADS_initialize
     PUBLIC :: ADS_update_horizontal_vcomps
     PUBLIC :: ADS_release
     PUBLIC :: ADS_solve_along_x
     PUBLIC :: ADS_solve_along_y
     PUBLIC :: ADS_solve_along_z
#if defined WITH_ACC
     PUBLIC :: ADS_get_vcomps_gpu
     PUBLIC :: ADS_get_diffusion
     PUBLIC :: ADS_update_diffusion
#endif
!
!    LIST OF PRIVATE ROUTINES IN THE MODULE
!
     PRIVATE :: KT_RHS
     PRIVATE :: r_minmod
     PRIVATE :: r_sbee
     PRIVATE :: r_ospre
     PRIVATE :: freeflow
     PRIVATE :: dirichlet
     PRIVATE :: periodic
#if defined WITH_ACC
     PRIVATE :: r_sbee2
#endif
!
!    Interface of abstract boundary conditions function
!
     abstract interface
     subroutine bconditions (c,mydim)
       import
       real(rp),   intent(INOUT) :: c  (my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h,my_kps_2h:my_kpe_2h)
       integer(ip),intent(IN   ) :: mydim
     end subroutine bconditions
     end interface

!
!    Interface of flux limiter function
!
     abstract interface
       function fluxlimiter (c,ips,ipe)
         import
         real(rp)                  :: fluxlimiter(ips-2:ipe+2)
         real(rp),   intent(IN)    :: c(ips-2:ipe+2)
         integer(ip),intent(IN)    :: ips,ipe
       end function fluxlimiter
     end interface

     CONTAINS
!
!-----------------------------------
!    subroutine ADS_initialize
!-----------------------------------
!
!>   @details
!>   Allocate resources and initialize parameters for ADS solver.
!
#if defined WITH_ACC
     subroutine ADS_initialize(limitertype,time_marching, MY_GRID,nbins)
#else
     subroutine ADS_initialize(limitertype,time_marching)
#endif
     implicit none
!
!>   @param limitertype    flux limiter flag       : MINMOD = 1, SUPERBEE = 2, OSPRE = 3
!>   @param time_marching  time integration scheme : TIME_MARCHING_EULER = 1, TIME_MARCHING_RK  = 2
!>   @param MY_GRID
!>   @param nbins
!
     integer(ip), intent(IN)    :: limitertype
     integer(ip), intent(IN)    :: time_marching
#if defined WITH_ACC
     type(ARAKAWA_C_GRID),intent(IN   ) :: MY_GRID
     integer(ip), intent(IN)    :: nbins
     integer(ip)                :: i,j,k,ibin
#else
     integer(ip)                :: gsize
#endif
     logical :: ads_stat
!
!*** Temporary velocity for current time
!
#if defined WITH_ACC
     allocate(ADS_u(my_ibs_1h:my_ibe_1h,my_jps   :my_jpe   ,my_kps   :my_kpe   ),pinned=ads_stat)
     if( .not. ads_stat ) print*, "ERROR: ADS_u was not pinned"
     allocate(ADS_v(my_jbs_1h:my_jbe_1h,my_ips   :my_ipe   ,my_kps   :my_kpe   ),pinned=ads_stat)
     if( .not. ads_stat ) print*, "ERROR: ADS_v was not pinned"
     allocate(ADS_w(my_kbs_1h:my_kbe_1h,my_ips   :my_ipe   ,my_jps   :my_jpe,nbins   ),pinned=ads_stat)
     if( .not. ads_stat ) print*, "ERROR: ADS_w was not pinned"
     allocate(ADS_k1 (my_ips_2h:my_ipe_2h,my_jps   :my_jpe   ,my_kps   :my_kpe   ),pinned=ads_stat)
     if( .not. ads_stat ) print*, "ERROR: ADS_k1 was not pinned"
     allocate(ADS_k2 (my_jps_2h:my_jpe_2h,my_ips   :my_ipe   ,my_kps   :my_kpe   ),pinned=ads_stat)
     if( .not. ads_stat ) print*, "ERROR: ADS_k2 was not pinned"
     allocate(ADS_k3 (my_kps_2h:my_kpe_2h,my_ips   :my_ipe   ,my_jps   :my_jpe   ),pinned=ads_stat)
     if( .not. ads_stat ) print*, "ERROR: ADS_k3 was not pinned"
#else
    allocate(CB%u(my_ibs_1h:my_ibe_1h,my_jps   :my_jpe   ,my_kps   :my_kpe   ))
    allocate(CB%v(my_jbs_1h:my_jbe_1h,my_ips   :my_ipe   ,my_kps   :my_kpe   ))
    allocate(CB%w(my_kbs_1h:my_kbe_1h,my_ips   :my_ipe   ,my_jps   :my_jpe   ))
#endif
!
!*** Configure boundary conditions
!
     CB%apply_bconditions => freeflow
     CB%bvalue(:) = 0.0_rp
#if defined WITH_ACC
     !$acc enter data copyin(CB)
     !$acc enter data copyin(CB%bvalue)
#endif
!
!*** Limiter
!
     select case(limitertype)
     case(1)
        !                         minmod
        CB%flux_limiter => r_minmod
     case(2)
        !                         superbee
        CB%flux_limiter => r_sbee
     case(3)
        !                         ospre
        CB%flux_limiter => r_ospre
     end select
#if defined WITH_ACC
     CB%limitertype = limitertype
#endif
!
!*** Time marching
!
     CB%time_marching = time_marching

#if defined WITH_ACC
!
!*** Resources for solvers
!
     allocate(xnewc(my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h,my_kps_2h:my_kpe_2h))
     allocate(xc0(my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h,my_kps_2h:my_kpe_2h))
     allocate(ynewc(my_jps_2h:my_jpe_2h,my_ips_2h:my_ipe_2h,my_kps_2h:my_kpe_2h))
     allocate(yc0(my_jps_2h:my_jpe_2h,my_ips_2h:my_ipe_2h,my_kps_2h:my_kpe_2h))
     allocate(yinic(my_jps_2h:my_jpe_2h,my_ips_2h:my_ipe_2h,my_kps_2h:my_kpe_2h))
     allocate(znewc(my_kps_2h:my_kpe_2h,my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h))
     allocate(zc0(my_kps_2h:my_kpe_2h,my_ips_2h:my_ipe_2h, my_jps_2h:my_jpe_2h))
     allocate(zinic(my_kps_2h:my_kpe_2h,my_ips_2h:my_ipe_2h, my_jps_2h:my_jpe_2h))
     allocate(xki(my_ips:my_ipe))
     allocate(yki(my_jps:my_jpe))
     allocate(zki(my_kps:my_kpe))
     allocate(ri(my_ips_2h:my_ipe_2h))
     allocate(rj(my_jps_2h:my_jpe_2h))
     allocate(rk(my_kps_2h:my_kpe_2h))
     allocate(Wswapi(1:2,my_jps:my_jpe,my_kps:my_kpe))
     allocate(Wswapo(1:2,my_jps:my_jpe,my_kps:my_kpe))
     allocate(Eswapi(1:2,my_jps:my_jpe,my_kps:my_kpe))
     allocate(Eswapo(1:2,my_jps:my_jpe,my_kps:my_kpe))
     allocate(Sswapi(1:2,my_ips:my_ipe,my_kps:my_kpe))
     allocate(Sswapo(1:2,my_ips:my_ipe,my_kps:my_kpe))
     allocate(Nswapi(1:2,my_ips:my_ipe,my_kps:my_kpe))
     allocate(Nswapo(1:2,my_ips:my_ipe,my_kps:my_kpe))
     allocate(Dswapi(1:2,my_ips:my_ipe,my_jps:my_jpe))
     allocate(Dswapo(1:2,my_ips:my_ipe,my_jps:my_jpe))
     allocate(Uswapi(1:2,my_ips:my_ipe,my_jps:my_jpe))
     allocate(Uswapo(1:2,my_ips:my_ipe,my_jps:my_jpe))
     allocate(s_kflux(my_kps:my_kpe))
     allocate(e_kflux(my_kps:my_kpe))
     allocate(s_iflux(my_ips:my_ipe))
     allocate(e_iflux(my_ips:my_ipe))
     allocate(dX1_p(my_ips_2h:my_ipe_2h))
     allocate(dX1_b(my_ibs_1h:my_ibe_1h))
     allocate(dX2_p(my_jps_2h:my_jpe_2h))
     allocate(dX2_b(my_jbs_1h:my_jbe_1h))
     allocate(dX3_p(my_kps_2h:my_kpe_2h))
     allocate(dX3_b(my_kbs_1h:my_kbe_1h))
     dX1_p(:)=MY_GRID%dX1_p(:)
     dX1_b(:)=MY_GRID%dX1_b(:)
     dX2_p(:)=MY_GRID%dX2_p(:)
     dX2_b(:)=MY_GRID%dX2_b(:)
     dX3_p(:)=MY_GRID%dX3_p(:)
     dX3_b(:)=MY_GRID%dX3_b(:)
!
!    Setting devices
!
     !$acc enter data copyin( dX1_b,dX1_p,dX2_b,dX2_p,dX3_b,dX3_p)
     !$acc enter data create(ADS_u,ADS_v,ADS_w,ADS_k1,ADS_k2,ADS_k3)
     !$acc enter data create(s_kflux,e_kflux,s_iflux,e_iflux,xc0,xnewc,yc0,ynewc,yinic,zc0,znewc,zinic)
     !$acc enter data create(Wswapi,Wswapo,Eswapi,Eswapo)
     !$acc enter data create(Sswapi,Sswapo,Nswapi,Nswapo)
     !$acc enter data create(Dswapi,Dswapo,Uswapi,Uswapo)
!
!*** Initialize ADS_w as it will be used
!

     !$acc parallel loop gang vector collapse(4)  default(none) present(ADS_w)
     do ibin = 1,nbins
       do j = my_jps,my_jpe
         do i = my_ips,my_ipe
           do k = my_kbs_1h,my_kbe_1h
             ADS_w(k,i,j,ibin) = 0.0_rp
           end do
         end do
       end do
     end do
!
#else
    !
    !*** Resources for KT_RHS
    !
    gsize = my_ipe - my_ips + 1
    if (gsize.lt.(my_jpe-my_jps+1)) then
       gsize = my_jpe-my_jps+1
    end if
    if (gsize.lt.(my_kpe-my_kps+1)) then
       gsize = my_kpe-my_kps+1
    end if

    allocate(CB%raw_r(gsize+4))
    allocate(CB%raw_c_r(gsize+2))
    allocate(CB%raw_c_l(gsize+2))

    gsize = (my_ipe_2h-my_ips_2h+1)*(my_jpe_2h-my_jps_2h+1)*(my_kpe_2h-my_kps_2h+1)

    allocate(CB%c0(gsize))
    allocate(CB%new_c(gsize))
    allocate(CB%raw_c(gsize))
#endif

     end subroutine ADS_initialize
!
!----------------------------------------
!    subroutine ADS_get_diffusion
!----------------------------------------
!
!>   @details
!>   Get pointers to diffusion components.
!
     subroutine ADS_get_diffusion(my_k1,my_k2,my_k3)
     implicit none
     real(rp),pointer,intent(INOUT) :: my_k1(:,:,:)
     real(rp),pointer,intent(INOUT) :: my_k2(:,:,:)
     real(rp),pointer,intent(INOUT) :: my_k3(:,:,:)

     my_k1 => ADS_k1
     my_k2 => ADS_k2
     my_k3 => ADS_k3

     end subroutine ADS_get_diffusion
!
!----------------------------------------
!    subroutine ADS_update_diffusion
!----------------------------------------
!
!>   @details
!>   Update diffusion components in device.
!
     subroutine ADS_update_diffusion
     implicit none

     !$acc update device(ADS_k1)
     !$acc update device(ADS_k2)
     !$acc update device(ADS_k3)

     end subroutine ADS_update_diffusion
!
!----------------------------------------
!    subroutine ADS_get_horizontal_vcomps
!----------------------------------------
!
!>   @details
!>   Get pointers to horizontal velocity components.
!
     subroutine ADS_get_vcomps_gpu(my_u,my_v,my_w)
     implicit none

     real(rp),pointer,intent(INOUT) :: my_u(:,:,:)
     real(rp),pointer,intent(INOUT) :: my_v(:,:,:)
#if defined WITH_ACC
     real(rp),pointer,intent(INOUT) :: my_w(:,:,:,:)
#else
     real(rp),pointer,intent(INOUT) :: my_w(:,:,:)
#endif

#if defined WITH_ACC
     my_u => ADS_u
     my_v => ADS_v
     my_w => ADS_w
#else
   my_u => CB%u
   my_v => CB%v
   my_w => CB%w
#endif

     end subroutine ADS_get_vcomps_gpu
!
!----------------------------------------
!    subroutine ADS_update_horizontal_vcomps
!----------------------------------------
!
!>   @details
!>   Update velocity horizontal components in device
!
#if defined WITH_ACC
     subroutine ADS_update_horizontal_vcomps()
#else
  subroutine ADS_update_horizontal_vcomps(my_u1,my_u2,my_v1,my_v2,stime)
#endif
     implicit none

#if defined WITH_ACC
!!     !$acc update device(ADS_u)
!!     !$acc update device(ADS_v)
#else
    real(rp),            intent(IN   ) :: my_u1 (my_ibs_1h:my_ibe_1h,my_jps   :my_jpe   ,my_kps   :my_kpe   )
    real(rp),            intent(IN   ) :: my_u2 (my_ibs_1h:my_ibe_1h,my_jps   :my_jpe   ,my_kps   :my_kpe   )
    real(rp),            intent(IN   ) :: my_v1 (my_jbs_1h:my_jbe_1h,my_ips   :my_ipe   ,my_kps   :my_kpe   )
    real(rp),            intent(IN   ) :: my_v2 (my_jbs_1h:my_jbe_1h,my_ips   :my_ipe   ,my_kps   :my_kpe   )
    real(rp),            intent(IN   ) :: stime

    CB%u(:,:,:) = (1.0_rp-stime) * my_u1(:,:,:) + stime * my_u2(:,:,:)
    CB%v(:,:,:) = (1.0_rp-stime) * my_v1(:,:,:) + stime * my_v2(:,:,:)
#endif

     end subroutine ADS_update_horizontal_vcomps

!
!-----------------------------------
!    subroutine ADS_release
!-----------------------------------
!
!>   @details
!>   Release resources used by ADS solver.
!
     subroutine ADS_release()
     implicit none

#if defined WITH_ACC
!
!*** Delete resource in device
!
     !$acc exit data delete(ADS_k1,ADS_k2,ADS_k3,ADS_u,ADS_v,ADS_w,dX1_b,dX1_p,dX3_b,dX2_b,dX2_p,dX3_p)
     !$acc exit data delete(s_kflux,e_kflux,s_iflux,e_iflux,xnewc,xc0,ynewc,yc0,yinic,znewc,zc0,zinic)
     !$acc exit data delete(Wswapi,Wswapo,Eswapi,Eswapo)
     !$acc exit data delete(Sswapi,Sswapo,Nswapi,Nswapo)
     !$acc exit data delete(Dswapi,Dswapo,Uswapi,Uswapo)
!
!*** Temporary velocity for current time
!
     deallocate(ADS_u)
     deallocate(ADS_v)
     deallocate(ADS_w)
     deallocate(ADS_k1)
     deallocate(ADS_k2)
     deallocate(ADS_k3)

     deallocate(xnewc)
     deallocate(xc0)
     deallocate(ynewc)
     deallocate(yc0)
     deallocate(yinic)
     deallocate(znewc)
     deallocate(zc0)
     deallocate(zinic)
     deallocate(xki)
     deallocate(yki)
     deallocate(zki)
     deallocate(Wswapi)
     deallocate(Wswapo)
     deallocate(Eswapi)
     deallocate(Eswapo)
     deallocate(Sswapi)
     deallocate(Sswapo)
     deallocate(Nswapi)
     deallocate(Nswapo)
     deallocate(Dswapi)
     deallocate(Dswapo)
     deallocate(Uswapi)
     deallocate(Uswapo)
     deallocate(ri)
     deallocate(rj)
     deallocate(rk)
     deallocate(dX1_p)
     deallocate(dX1_b)
     deallocate(dX2_p)
     deallocate(dX2_b)
     deallocate(dX3_p)
     deallocate(dX3_b)
     deallocate(s_kflux)
     deallocate(e_kflux)
     deallocate(s_iflux)
     deallocate(e_iflux)
     !$acc exit data delete(CB%bvalue,CB)
#else
    deallocate(CB%u)
    deallocate(CB%v)
    deallocate(CB%w)
    deallocate(CB%raw_r)
    deallocate(CB%raw_c_r)
    deallocate(CB%raw_c_l)
    deallocate(CB%raw_c)
    deallocate(CB%new_c)
    deallocate(CB%c0)
#endif


     end subroutine ADS_release
!
!-----------------------------------
!    subroutine ADS_solve_along_x
!-----------------------------------
!
!>   @details
!>   Solves a time step dt of the 1D ADS equation along x direction.
!
#if defined WITH_ACC
  subroutine ADS_solve_along_x(my_W_flux,my_E_flux,dt,my_c,my_k1,nbins)
#else
  subroutine ADS_solve_along_x(my_W_flux,my_E_flux,dt,my_c,my_k1,MY_GRID,nbins)
#endif
     implicit none
!
!>   @param my_W_flux increment of mass flux across the W domain boundary during dt
!>   @param my_E_flux increment of mass flux across the E domain boundary during dt
!>   @param dt        time increment
!>   @param my_c      my values of scaled transported variable (concentration) at mass points
!>   @param my_k1     my values of scaled diffusion at mass points
    !>   @param MY_GRID   grid structure
!>   @param nbins     number of bins
!
     integer(ip),         intent(IN   ) :: nbins
     real(rp),            intent(INOUT) :: my_W_flux
     real(rp),            intent(INOUT) :: my_E_flux
     real(rp),            intent(IN   ) :: dt
#if defined WITH_ACC
     real(rp),            intent(INOUT) :: my_c(my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h,my_kps_2h:my_kpe_2h,nbins)
     real(rp),pointer,    intent(IN   ) :: my_k1 (:,:,:)
#else
     real(rp),target,     intent(INOUT) :: my_c(my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h,my_kps_2h:my_kpe_2h,nbins)
     type(ARAKAWA_C_GRID),intent(IN   ) :: MY_GRID
     real(rp),target,     intent(IN   ) :: my_k1 (my_ips_2h:my_ipe_2h,my_jps   :my_jpe   ,my_kps   :my_kpe   )
#endif

#if defined WITH_ACC
     integer(ip)           :: limtype,i,idx
     real(rp),allocatable  :: Wtflux(:,:),Etflux(:,:)
#else
    real(rp), pointer     :: my_c_1d(:)
    real(rp), pointer     :: my_u_1d(:)
    real(rp), pointer     :: my_k_1d(:)
    real(rp), pointer     :: c0(:,:,:)
    real(rp), pointer     :: new_c(:,:,:)
    real(rp), allocatable :: ki (:)
#endif
     integer(ip)           :: j,k,ibin
     real(rp)              :: flux(2),area

     !$acc data &
     !$acc present(xnewc,xc0,dX1_b,dX1_p,ADS_u,ADS_k1,dX3_b,dX2_b,Wswapi,Wswapo,Eswapi,Eswapo) &
     !$acc present(my_c(:,:,:,:)) &
     !$acc present(my_W_flux,my_E_flux) &
     !$acc present(CB) &
     !$acc present(CB%bvalue(:))

!*** Update concentration field halos
!
     do ibin = 1,nbins
       call domain_swap_mass_points_2halo_x_GPU(my_c(:,:,:,ibin))
     end do

#if defined WITH_ACC
     allocate(Wtflux(my_kps:my_kpe,my_jps:my_jpe))
     allocate(Etflux(my_kps:my_kpe,my_jps:my_jpe))
     !$acc data create(Wtflux,Etflux)
     !$acc kernels default(none)
     Wtflux(:,:) = 0.0_rp
     Etflux(:,:) = 0.0_rp
     !$acc end kernels
#endif
!
!*** Apply boundaries
!
     do ibin = 1,nbins
       call CB%apply_bconditions( my_c(:,:,:,ibin), 1_ip )
     end do

#if defined WITH_ACC
     limtype = CB%limitertype
#else
    !
    !*** Allocate memory
    !
    allocate(ki     (my_ips   :my_ipe   ))
    !
    !*** Prepare shape of temporary fields
    !
    c0(my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h, my_kps_2h:my_kpe_2h) => CB%c0(:)
    new_c(my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h, my_kps_2h:my_kpe_2h) => CB%new_c(:)
    !
#endif

! The main difference between CPU & GPU versions is that the "Loop over bins" is
! done outside (outer one) the select case in the former, whereas in the latter such loop is inserted inside the cases
#if defined WITH_ACC
     !
     select case(CB%time_marching)
     !
     case(TIME_MARCHING_EULER)
     !
     !   Euler O(1)
     !
!
!*** Loop over bins
!
     do ibin = 1,nbins
       !$acc parallel loop independent gang worker &
       !$acc private(ri,xki,flux,k,j,i,area)
       do idx = 1, (my_jpe-my_jps+1)*(my_kpe-my_kps+1)
         j = my_jps + mod(idx-1,my_jpe-my_jps+1)
         k = my_kps + (idx-1) / (my_jpe-my_jps+1)

         call KT_RHS(xki,my_c(my_ips_2h:my_ipe_2h,j,k,ibin), &
                 dX1_p(my_ips_2h:my_ipe_2h), &
                 dX1_b(my_ibs_1h:my_ibe_1h), &
                 ADS_u (my_ibs_1h:my_ibe_1h,j,k), &
                 ADS_k1(my_ips_2h:my_ipe_2h,j,k), &
                 flux,my_ips,my_ipe,my_ibs,my_ibe, &
                 ri, limtype)

         !$acc loop vector
         do i = my_ips,my_ipe
           my_c(i,j,k,ibin) = my_c(i,j,k,ibin) + dt*xki(i)
         end do

         !
         !*** Fluxes
         !
         area = dX3_b(k)*dX2_b(j)
         if(my_W_proc.eq.-1) Wtflux(k,j) = Wtflux(k,j) + flux(1) *area*dt
         if(my_E_proc.eq.-1) Etflux(k,j) = Etflux(k,j) + flux(2) *area*dt
         !
       end do
       !$acc end parallel
     end do
     !$acc parallel loop collapse(2) default(none)  &
     !$acc reduction(+:my_W_flux,my_E_flux)
     do j = my_jps,my_jpe
     do k = my_kps, my_kpe
       my_W_flux = my_W_flux + Wtflux(k,j)
       my_E_flux = my_E_flux + Etflux(k,j)
     end do
     end do

       !
     case(TIME_MARCHING_RK)
       !
       !   RK O(4)
       !

       ! openACC collapse directive does not work properly, we get numerical
       ! errors. So, in meantime single compacted loop is used.

!
!*** Loop over bins
!
     do ibin = 1,nbins

       !$acc parallel num_workers(16)
       !$acc loop independent gang worker private(ri,xki,flux,k,j)
       do idx = 1, (my_jpe-my_jps+1)*(my_kpe-my_kps+1)
         j = my_jps + mod(idx-1,my_jpe-my_jps+1)
         k = my_kps + (idx-1) / (my_jpe-my_jps+1)
!
!***     Time integration in the interval (t,t+dt) using a RK4 Runge-Kutta method
!
         call KT_RHS(xki,my_c(my_ips_2h:my_ipe_2h,j,k,ibin), &
                 dX1_p(my_ips_2h:my_ipe_2h), &
                 dX1_b(my_ibs_1h:my_ibe_1h), &
                 ADS_u (my_ibs_1h:my_ibe_1h,j,k), &
                 ADS_k1(my_ips_2h:my_ipe_2h,j,k), &
                 flux,my_ips,my_ipe,my_ibs,my_ibe, &
                 ri, limtype)

         ! In this first stage, we need to copy also the boundary halos.
         !$acc loop vector
         do i = my_ips_2h,my_ips-1
           xc0(i,j,k) = my_c(i,j,k,ibin)
           xnewc(i,j,k) = my_c(i,j,k,ibin)
         end do
         !$acc loop vector
         do i = my_ipe+1,my_ipe_2h
           xc0(i,j,k) = my_c(i,j,k,ibin)
           xnewc(i,j,k) = my_c(i,j,k,ibin)
         end do

         !$acc loop vector
         do i = my_ips,my_ipe
           xc0(i,j,k) = my_c(i,j,k,ibin) + 0.5_rp*dt*xki(i)

           xnewc(i,j,k) = my_c(i,j,k,ibin) + dt*xki(i) / 6.0_rp
         end do

         if (my_W_proc.ne.-1) Wswapo(1:2,j,k) = xc0(my_ips:my_ips+1,j,k)
         if (my_E_proc.ne.-1) Eswapo(1:2,j,k) = xc0(my_ipe-1:my_ipe,j,k)

       end do
       !$acc end parallel

        call domain_swap_mass_points_2halo_first_GPU (Wswapi, Wswapo, Eswapi,Eswapo, &
            my_jps,my_jpe,my_kps,my_kpe, my_W_proc, my_E_proc)

       !$acc parallel loop independent gang worker private(ri,xki,flux,k,j)
       do idx = 1, (my_jpe-my_jps+1)*(my_kpe-my_kps+1)
         j = my_jps + mod(idx-1,my_jpe-my_jps+1)
         k = my_kps + (idx-1) / (my_jpe-my_jps+1)

         if (my_W_proc.ne.-1) xc0(my_ips-2:my_ips-1,j,k) = Wswapi(1:2,j,k)
         if (my_E_proc.ne.-1) xc0(my_ipe+1:my_ipe+2,j,k) = Eswapi(1:2,j,k)

         call KT_RHS(xki,xc0(my_ips_2h:my_ipe_2h,j,k), &
                 dX1_p(my_ips_2h:my_ipe_2h), &
                 dX1_b(my_ibs_1h:my_ibe_1h), &
                 ADS_u (my_ibs_1h:my_ibe_1h,j,k), &
                 ADS_k1(my_ips_2h:my_ipe_2h,j,k), &
                 flux,my_ips,my_ipe,my_ibs,my_ibe, &
                 ri, limtype)

         !$acc loop vector
         do i = my_ips,my_ipe
           xc0(i,j,k) = my_c(i,j,k,ibin) + 0.5_rp*dt*xki(i)

           xnewc(i,j,k) = xnewc(i,j,k) + dt*xki(i) / 3.0_rp
         end do

         if (my_W_proc.ne.-1) Wswapo(1:2,j,k) = xc0(my_ips:my_ips+1,j,k)
         if (my_E_proc.ne.-1) Eswapo(1:2,j,k) = xc0(my_ipe-1:my_ipe,j,k)

       end do
       !$acc end parallel

       call domain_swap_mass_points_2halo_first_GPU (Wswapi, Wswapo, Eswapi,Eswapo, &
            my_jps,my_jpe,my_kps,my_kpe, my_W_proc, my_E_proc)

       !$acc parallel loop gang worker &
       !$acc private(ri,xki,flux,k,j)
       do idx = 1, (my_jpe-my_jps+1)*(my_kpe-my_kps+1)
         j = my_jps + mod(idx-1,my_jpe-my_jps+1)
         k = my_kps + (idx-1) / (my_jpe-my_jps+1)

         if (my_W_proc.ne.-1) xc0(my_ips-2:my_ips-1,j,k) = Wswapi(1:2,j,k)
         if (my_E_proc.ne.-1) xc0(my_ipe+1:my_ipe+2,j,k) = Eswapi(1:2,j,k)

         call KT_RHS(xki,xc0(my_ips_2h:my_ipe_2h,j,k), &
                 dX1_p(my_ips_2h:my_ipe_2h), &
                 dX1_b(my_ibs_1h:my_ibe_1h), &
                 ADS_u (my_ibs_1h:my_ibe_1h,j,k), &
                 ADS_k1(my_ips_2h:my_ipe_2h,j,k), &
                 flux,my_ips,my_ipe,my_ibs,my_ibe, &
                 ri, limtype)

         !$acc loop vector
         do i = my_ips,my_ipe
           xc0(i,j,k) = my_c(i,j,k,ibin) + dt*xki(i)

           xnewc(i,j,k) = xnewc(i,j,k) + dt*xki(i) / 3.0_rp
         end do

         if (my_W_proc.ne.-1) Wswapo(1:2,j,k) = xc0(my_ips:my_ips+1,j,k)
         if (my_E_proc.ne.-1) Eswapo(1:2,j,k) = xc0(my_ipe-1:my_ipe,j,k)

       end do
       !$acc end parallel

       call domain_swap_mass_points_2halo_first_GPU (Wswapi, Wswapo, Eswapi,Eswapo, &
            my_jps,my_jpe,my_kps,my_kpe, my_W_proc, my_E_proc)

       !$acc parallel loop independent gang worker &
       !$acc private(ri,xki,flux,k,j,i,area)
       do idx = 1, (my_jpe-my_jps+1)*(my_kpe-my_kps+1)
         j = my_jps + mod(idx-1,my_jpe-my_jps+1)
         k = my_kps + (idx-1) / (my_jpe-my_jps+1)

         if (my_W_proc.ne.-1) xc0(my_ips-2:my_ips-1,j,k) = Wswapi(1:2,j,k)
         if (my_E_proc.ne.-1) xc0(my_ipe+1:my_ipe+2,j,k) = Eswapi(1:2,j,k)

         call KT_RHS(xki,xc0(my_ips_2h:my_ipe_2h,j,k), &
                 dX1_p(my_ips_2h:my_ipe_2h), &
                 dX1_b(my_ibs_1h:my_ibe_1h), &
                 ADS_u (my_ibs_1h:my_ibe_1h,j,k), &
                 ADS_k1(my_ips_2h:my_ipe_2h,j,k), &
                 flux,my_ips,my_ipe,my_ibs,my_ibe, &
                 ri, limtype)

         !$acc loop vector
         do i = my_ips,my_ipe
           my_c(i,j,k,ibin) = xnewc(i,j,k) + dt*xki(i) / 6.0_rp
         end do
         !
         !*** Fluxes
         !
         area = dX3_b(k)*dX2_b(j)
         if(my_W_proc.eq.-1) Wtflux(k,j) = Wtflux(k,j) + flux(1) *area*dt
         if(my_E_proc.eq.-1) Etflux(k,j) = Etflux(k,j) + flux(2) *area*dt
         !
       end do
       !$acc end parallel
     end do
     !$acc parallel loop collapse(2) present(my_W_flux,my_E_flux,Wtflux,Etflux) &
     !$acc reduction(+:my_W_flux,my_E_flux)
     do j = my_jps,my_jpe
     do k = my_kps, my_kpe
       my_W_flux = my_W_flux + Wtflux(k,j)
       my_E_flux = my_E_flux + Etflux(k,j)
     end do
     end do

     end select
     !$acc end data
     deallocate(Wtflux)
     deallocate(Etflux)
     !
!     do ibin = 1,nbins
!       call domain_swap_mass_points_2halo_x_GPU(my_c(:,:,:,ibin))
!     end do
     !$acc end data
#else
    !
    !*** Loop over bins
    !
    do ibin = 1,nbins
       !
       select case(CB%time_marching)
          !
       case(TIME_MARCHING_EULER)
          !
          !   Euler O(1)
          !
          do k = my_kps,my_kpe
             do j = my_jps,my_jpe
                !
                my_c_1d(my_ips_2h:my_ipe_2h) => my_c (my_ips_2h:my_ipe_2h,j,k,ibin)
                my_k_1d(my_ips_2h:my_ipe_2h) => my_k1(my_ips_2h:my_ipe_2h,j,k)
                my_u_1d(my_ibs_1h:my_ibe_1h) => CB%u (my_ibs_1h:my_ibe_1h,j,k)
                !
                ki = KT_RHS(my_c_1d,MY_GRID%dX1_p,MY_GRID%dX1_b,my_u_1d,my_k_1d, &
                     flux,my_ips,my_ipe,my_ibs,my_ibe)
                !
                my_c(my_ips:my_ipe,j,k,ibin) = my_c_1d(my_ips:my_ipe) + dt*ki(my_ips:my_ipe)
                !
                !*** Fluxes
                !
                area = MY_GRID%dX3_b(k)*MY_GRID%dX2_b(j)
                if(my_W_proc.eq.-1) my_W_flux = my_W_flux + flux(1) *area*dt
                if(my_E_proc.eq.-1) my_E_flux = my_E_flux + flux(2) *area*dt
                !
             end do
          end do
          !
       case(TIME_MARCHING_RK)
          !
          !   RK O(4)
          !
          do k = my_kps,my_kpe
             do j = my_jps,my_jpe
                !
                my_c_1d(my_ips_2h:my_ipe_2h) => my_c (my_ips_2h:my_ipe_2h,j,k,ibin)
                my_k_1d(my_ips_2h:my_ipe_2h) => my_k1(my_ips_2h:my_ipe_2h,j,k)
                my_u_1d(my_ibs_1h:my_ibe_1h) => CB%u (my_ibs_1h:my_ibe_1h,j,k)
                !
                !***     Time integration in the interval (t,t+dt) using a RK4 Runge-Kutta method
                !
                ki = KT_RHS(my_c_1d,MY_GRID%dX1_p,MY_GRID%dX1_b,my_u_1d,my_k_1d, &
                     flux,my_ips,my_ipe,my_ibs,my_ibe)

                ! In this first stage, we need to copy also the boundary halos.
                c0(:,j,k) = my_c_1d(:)
                c0(my_ips:my_ipe,j,k) = c0(my_ips:my_ipe,j,k) + 0.5_rp*dt*ki(my_ips:my_ipe)

                new_c(:,j,k) = my_c_1d(:)
                new_c(my_ips:my_ipe,j,k) = new_c(my_ips:my_ipe,j,k) + dt*ki(my_ips:my_ipe) / 6.0_rp

             end do
          end do
          !
          !      Successive approaches. Note that swap is necessary to update predictor halo values
          !
          call domain_swap_mass_points_2halo_first (c0(:,:,:), my_ips, my_ipe, &
               my_jps,my_jpe,my_kps,my_kpe, my_W_proc, my_E_proc)

          do k = my_kps,my_kpe
             do j = my_jps,my_jpe
                !
                my_c_1d(my_ips_2h:my_ipe_2h) => c0(my_ips_2h:my_ipe_2h,j,k)
                my_k_1d(my_ips_2h:my_ipe_2h) => my_k1(my_ips_2h:my_ipe_2h,j,k)
                my_u_1d(my_ibs_1h:my_ibe_1h) => CB%u (my_ibs_1h:my_ibe_1h,j,k)
                !
                ki = KT_RHS(my_c_1d,MY_GRID%dX1_p,MY_GRID%dX1_b,my_u_1d,my_k_1d, &
                     flux,my_ips,my_ipe,my_ibs,my_ibe)
                !
                c0(my_ips:my_ipe,j,k) = my_c(my_ips:my_ipe,j,k,ibin) + 0.5_rp*dt*ki(my_ips:my_ipe)
                new_c(my_ips:my_ipe,j,k) = new_c(my_ips:my_ipe,j,k) + dt*ki(my_ips:my_ipe) / 3.0_rp
             end do
          end do
          !
          call domain_swap_mass_points_2halo_first (c0(:,:,:), my_ips, my_ipe, &
               my_jps,my_jpe,my_kps,my_kpe, my_W_proc, my_E_proc)

          do k = my_kps,my_kpe
             do j = my_jps,my_jpe
                !
                my_c_1d(my_ips_2h:my_ipe_2h) => c0(my_ips_2h:my_ipe_2h,j,k)
                my_k_1d(my_ips_2h:my_ipe_2h) => my_k1(my_ips_2h:my_ipe_2h,j,k)
                my_u_1d(my_ibs_1h:my_ibe_1h) => CB%u (my_ibs_1h:my_ibe_1h,j,k)

                ki = KT_RHS(my_c_1d,MY_GRID%dX1_p,MY_GRID%dX1_b,my_u_1d,my_k_1d, &
                     flux,my_ips,my_ipe,my_ibs,my_ibe)
                !
                c0(my_ips:my_ipe,j,k) = my_c(my_ips:my_ipe,j,k,ibin) + dt*ki(my_ips:my_ipe)
                new_c(my_ips:my_ipe,j,k) = new_c(my_ips:my_ipe,j,k) + dt*ki(my_ips:my_ipe) / 3.0_rp
             end do
          end do

          call domain_swap_mass_points_2halo_first (c0(:,:,:), my_ips,my_ipe, &
               my_jps,my_jpe,my_kps,my_kpe, my_W_proc, my_E_proc)

          do k = my_kps,my_kpe
             do j = my_jps,my_jpe
                !
                my_c_1d(my_ips_2h:my_ipe_2h) => c0(my_ips_2h:my_ipe_2h,j,k)
                my_k_1d(my_ips_2h:my_ipe_2h) => my_k1(my_ips_2h:my_ipe_2h,j,k)
                my_u_1d(my_ibs_1h:my_ibe_1h) => CB%u (my_ibs_1h:my_ibe_1h,j,k)

                ki = KT_RHS(my_c_1d,MY_GRID%dX1_p,MY_GRID%dX1_b,my_u_1d,my_k_1d, &
                     flux,my_ips,my_ipe,my_ibs,my_ibe)
                !
                my_c(my_ips:my_ipe,j,k,ibin) = new_c(my_ips:my_ipe,j,k) + dt*ki(my_ips:my_ipe) / 6.0_rp
                !
                !*** Fluxes
                !
                area = MY_GRID%dX3_b(k)*MY_GRID%dX2_b(j)
                if(my_W_proc.eq.-1) my_W_flux = my_W_flux + flux(1) *area*dt
                if(my_E_proc.eq.-1) my_E_flux = my_E_flux + flux(2) *area*dt
                !
             end do
          end do
          !
          !
       end select
       !
    end do
    !
    deallocate(ki )
    !
!    do ibin = 1,nbins
!      call domain_swap_mass_points_2halo_x (my_c(:,:,:,ibin))
!    end do
#endif

     return
     end subroutine ADS_solve_along_x
!
!
!-----------------------------------
!    subroutine ADS_solve_along_y
!-----------------------------------
!
!>   @details
!>   Solves a time step dt of the 1D ADS equation along y direction.
!
#if defined WITH_ACC
     subroutine ADS_solve_along_y(my_S_flux,my_N_flux,dt,my_c,nbins)
#else
  subroutine ADS_solve_along_y(my_S_flux,my_N_flux,dt,my_c,MY_GRID,nbins)
#endif
     implicit none
!
!>   @param my_S_flux increment of mass flux across the S domain boundary during dt
!>   @param my_N_flux increment of mass flux across the N domain boundary during dt
!>   @param dt        time increment
!>   @param my_c      my values of scaled transported variable (concentration) at mass points
!>   @param my_k2     my values of scaled diffusion at mass points
!>   @param nbins     number of bins
    !>   @param MY_GRID   grid structure
!
     integer(ip),         intent(IN   ) :: nbins
     real(rp),            intent(INOUT) :: my_S_flux
     real(rp),            intent(INOUT) :: my_N_flux
     real(rp),            intent(IN   ) :: dt
#if defined WITH_ACC
     real(rp),            intent(INOUT) :: my_c  (my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h,my_kps_2h:my_kpe_2h,nbins)
#else
     real(rp),            intent(INOUT) :: my_c  (my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h,my_kps_2h:my_kpe_2h,nbins)
     real(rp),target     ,intent(IN   ) :: my_k2 (my_jps_2h:my_jpe_2h,my_ips   :my_ipe   ,my_kps   :my_kpe   )
    type(ARAKAWA_C_GRID),intent(IN   ) :: MY_GRID
#endif
#if defined WITH_ACC
     integer(ip)           :: limtype,j,idx
     real(rp),allocatable  :: Stflux(:,:),Ntflux(:,:)
#else
    real(rp), pointer     :: my_c_1d(:)
    real(rp), pointer     :: my_v_1d(:)
    real(rp), pointer     :: my_k_1d(:)
    real(rp), pointer     :: c0(:,:,:)
    real(rp), pointer     :: new_c(:,:,:)
    real(rp), pointer     :: ini_c(:,:,:)
    real(rp), allocatable :: ki (:)
#endif
     integer(ip)           :: i,k,ibin
     real(rp)              :: flux(2),area

     !$acc data &
     !$acc present(yinic,ynewc,yc0,dX2_b,dX2_p,ADS_v,ADS_k2,dX3_b,dX1_b,Sswapi,Sswapo,Nswapi,Nswapo) &
     !$acc present(my_c(:,:,:,:)) &
     !$acc present(my_S_flux,my_N_flux) &
     !$acc present(CB,CB%bvalue(:))
!
!*** Update concentration field halos
!
     do ibin = 1,nbins
       call domain_swap_mass_points_2halo_y_GPU (my_c(:,:,:,ibin))
     end do
!
!*** Apply boundaries
!

#if defined WITH_ACC
     allocate(Stflux(my_kps:my_kpe,my_ips:my_ipe))
     allocate(Ntflux(my_kps:my_kpe,my_ips:my_ipe))
     !$acc data create(Stflux,Ntflux)
     !$acc kernels default(none)
     Stflux(:,:) = 0.0_rp
     Ntflux(:,:) = 0.0_rp
     !$acc end kernels
#endif

     do ibin = 1, nbins
       call CB%apply_bconditions( my_c(:,:,:,ibin), 2_ip )
     end do

#if defined WITH_ACC
     limtype = CB%limitertype
#else
    !
    !*** Allocate memory
    !
    allocate(ki     (my_jps   :my_jpe   ))
    !
    !*** Prepare shape of temporary fields
    !
    c0(my_jps_2h:my_jpe_2h,my_ips_2h:my_ipe_2h, my_kps_2h:my_kpe_2h) => CB%c0(:)
    new_c(my_jps_2h:my_jpe_2h,my_ips_2h:my_ipe_2h, my_kps_2h:my_kpe_2h) => CB%new_c(:)
    ini_c(my_jps_2h:my_jpe_2h,my_ips_2h:my_ipe_2h, my_kps_2h:my_kpe_2h) => CB%raw_c(:)
    !
#endif

! The main difference between CPU & GPU versions is that the "Loop over bins" is
! done outside (outer one) the select case in the former, whereas in the latter such loop is inserted inside the cases
#if defined WITH_ACC
     !
     select case(CB%time_marching)
     !
     case(TIME_MARCHING_EULER)
       !
       !   Euler O(1)
       !
!
!*** Loop over bins
!
     do ibin = 1, nbins
       !$acc parallel loop independent gang worker num_workers(16) &
       !$acc private(rj,yki,flux,k,i,j,area)
       do idx = 1, (my_ipe-my_ips+1)*(my_kpe-my_kps+1)
         i = my_ips + mod(idx-1,my_ipe-my_ips+1)
         k = my_kps + (idx-1) / (my_ipe-my_ips+1)

         !$acc loop vector
         do j = my_jps_2h,my_jpe_2h
           yinic(j,i,k) = my_c(i,j,k,ibin)
         end do

         call KT_RHS(yki,yinic(my_jps_2h:my_jpe_2h,i,k), &
                     dX2_p(my_jps_2h:my_jpe_2h), &
                     dX2_b(my_jbs_1h:my_jbe_1h), &
                     ADS_v (my_jbs_1h:my_jbe_1h,i,k), &
                     ADS_k2(my_jps_2h:my_jpe_2h,i,k), &
                     flux,my_jps,my_jpe,my_jbs,my_jbe, &
                     rj, limtype)

         !$acc loop vector
         do j = my_jps,my_jpe
           my_c(i,j,k,ibin) = my_c(i,j,k,ibin) + dt*yki(j)
         end do
!
!***     Fluxes
!
         area = dX3_b(k)*dX1_b(i)
         if(my_S_proc.eq.-1) Stflux(k,i) = Stflux(k,i) + flux(1) *area*dt
         if(my_N_proc.eq.-1) Ntflux(k,i) = Ntflux(k,i) + flux(2) *area*dt
         !
       end do
       !$acc end parallel
     end do
     !$acc parallel loop collapse(2) default(none) &
     !$acc reduction(+:my_S_flux,my_N_flux)
     do i = my_ips,my_ipe
     do k = my_kps, my_kpe
       my_S_flux = my_S_flux + Stflux(k,i)
       my_N_flux = my_N_flux + Ntflux(k,i)
     end do
     end do

       !
     case(TIME_MARCHING_RK)
       !
       !   RK O(4)
       !
!
!*** Loop over bins
!
     do ibin = 1,nbins
       !$acc parallel loop independent gang worker num_workers(16) private(rj,yki,flux,k,i)
       do idx = 1, (my_ipe-my_ips+1)*(my_kpe-my_kps+1)
         i = my_ips + mod(idx-1,my_ipe-my_ips+1)
         k = my_kps + (idx-1) / (my_ipe-my_ips+1)

         ! In this first stage, we need to copy also the boundary halos
         ! in new_c and c0.
         !$acc loop vector
         do j = my_jps_2h,my_jpe_2h
           yinic(j,i,k) = my_c(i,j,k,ibin)
           ynewc(j,i,k) = my_c(i,j,k,ibin)
           yc0(j,i,k) = my_c(i,j,k,ibin)
         end do
!
!***     Time integration in the interval (t,t+dt) using a RK4 Runge-Kutta method
!
         call KT_RHS(yki,yinic(my_jps_2h:my_jpe_2h,i,k), &
                     dX2_p(my_jps_2h:my_jpe_2h), &
                     dX2_b(my_jbs_1h:my_jbe_1h), &
                     ADS_v (my_jbs_1h:my_jbe_1h,i,k), &
                     ADS_k2(my_jps_2h:my_jpe_2h,i,k), &
                     flux,my_jps,my_jpe,my_jbs,my_jbe, &
                     rj, limtype)

         !$acc loop vector
         do j = my_jps,my_jpe
           yc0(j,i,k) = yc0(j,i,k) + 0.5_rp*dt*yki(j)

           ynewc(j,i,k) = ynewc(j,i,k) + dt*yki(j) / 6.0_rp
         end do

         if (my_S_proc.ne.-1) Sswapo(1:2,i,k) = yc0(my_jps:my_jps+1,i,k)
         if (my_N_proc.ne.-1) Nswapo(1:2,i,k) = yc0(my_jpe-1:my_jpe,i,k)
       end do
       !$acc end parallel
!
!      Successive approaches. Note that swap is necessary to update predictor halo values
!
       call domain_swap_mass_points_2halo_first_GPU (Sswapi, Sswapo, Nswapi,Nswapo, &
           my_ips,my_ipe,my_kps,my_kpe, my_S_proc, my_N_proc)

       !$acc parallel loop independent gang worker num_workers(16) &
       !$acc private(rj,yki,flux,k,i)
       do idx = 1, (my_ipe-my_ips+1)*(my_kpe-my_kps+1)
         i = my_ips + mod(idx-1,my_ipe-my_ips+1)
         k = my_kps + (idx-1) / (my_ipe-my_ips+1)

         if (my_S_proc.ne.-1) yc0(my_jps-2:my_jps-1,i,k) = Sswapi(1:2,i,k)
         if (my_N_proc.ne.-1) yc0(my_jpe+1:my_jpe+2,i,k) = Nswapi(1:2,i,k)

         call KT_RHS(yki,yc0(my_jps_2h:my_jpe_2h,i,k), &
                     dX2_p(my_jps_2h:my_jpe_2h), &
                     dX2_b(my_jbs_1h:my_jbe_1h), &
                     ADS_v (my_jbs_1h:my_jbe_1h,i,k), &
                     ADS_k2(my_jps_2h:my_jpe_2h,i,k), &
                     flux,my_jps,my_jpe,my_jbs,my_jbe, &
                     rj, limtype)

         !$acc loop independent vector
         do j = my_jps,my_jpe
           yc0(j,i,k) = yinic(j,i,k) + 0.5_rp*dt*yki(j)
           ynewc(j,i,k) = ynewc(j,i,k) + dt*yki(j) / 3.0_rp
         end do

         if (my_S_proc.ne.-1) Sswapo(1:2,i,k) = yc0(my_jps:my_jps+1,i,k)
         if (my_N_proc.ne.-1) Nswapo(1:2,i,k) = yc0(my_jpe-1:my_jpe,i,k)
       end do
       !$acc end parallel

       call domain_swap_mass_points_2halo_first_GPU (Sswapi, Sswapo, Nswapi,Nswapo, &
           my_ips,my_ipe,my_kps,my_kpe, my_S_proc, my_N_proc)

       !$acc parallel loop independent gang worker num_workers(16) &
       !$acc private(rj,yki,flux,k,i)
       do idx = 1, (my_ipe-my_ips+1)*(my_kpe-my_kps+1)
         i = my_ips + mod(idx-1,my_ipe-my_ips+1)
         k = my_kps + (idx-1) / (my_ipe-my_ips+1)

         if (my_S_proc.ne.-1) yc0(my_jps-2:my_jps-1,i,k) = Sswapi(1:2,i,k)
         if (my_N_proc.ne.-1) yc0(my_jpe+1:my_jpe+2,i,k) = Nswapi(1:2,i,k)

         call KT_RHS(yki,yc0 (my_jps_2h:my_jpe_2h,i,k), &
                     dX2_p,dX2_b, &
                     ADS_v (my_jbs_1h:my_jbe_1h,i,k), &
                     ADS_k2(my_jps_2h:my_jpe_2h,i,k), &
                     flux,my_jps,my_jpe,my_jbs,my_jbe, &
                     rj, limtype)
!
         !$acc loop vector
         do j = my_jps,my_jpe
           yc0(j,i,k) = yinic(j,i,k) + dt*yki(j)
           ynewc(j,i,k) = ynewc(j,i,k) + dt*yki(j) / 3.0_rp
         end do

         if (my_S_proc.ne.-1) Sswapo(1:2,i,k) = yc0(my_jps:my_jps+1,i,k)
         if (my_N_proc.ne.-1) Nswapo(1:2,i,k) = yc0(my_jpe-1:my_jpe,i,k)
       end do
       !$acc end parallel

       call domain_swap_mass_points_2halo_first_GPU (Sswapi, Sswapo, Nswapi,Nswapo, &
           my_ips,my_ipe,my_kps,my_kpe, my_S_proc, my_N_proc)

       !$acc parallel loop independent gang worker num_workers(16) &
       !$acc private(rj,yki,flux,k,i,j,area)
       do idx = 1, (my_ipe-my_ips+1)*(my_kpe-my_kps+1)
         i = my_ips + mod(idx-1,my_ipe-my_ips+1)
         k = my_kps + (idx-1) / (my_ipe-my_ips+1)

         if (my_S_proc.ne.-1) yc0(my_jps-2:my_jps-1,i,k) = Sswapi(1:2,i,k)
         if (my_N_proc.ne.-1) yc0(my_jpe+1:my_jpe+2,i,k) = Nswapi(1:2,i,k)

         call KT_RHS(yki, yc0 (my_jps_2h:my_jpe_2h,i,k), &
                     dX2_p,dX2_b, &
                     ADS_v (my_jbs_1h:my_jbe_1h,i,k), &
                     ADS_k2(my_jps_2h:my_jpe_2h,i,k), &
                     flux,my_jps,my_jpe,my_jbs,my_jbe, &
                     rj, limtype)
!
         !$acc loop vector
         do j = my_jps,my_jpe
           my_c(i,j,k,ibin) = ynewc(j,i,k) + dt*yki(j) / 6.0_rp
         end do
!
!***     Fluxes
!
         area = dX3_b(k)*dX1_b(i)
         if(my_S_proc.eq.-1) Stflux(k,i) = Stflux(k,i) + flux(1) *area*dt
         if(my_N_proc.eq.-1) Ntflux(k,i) = Ntflux(k,i) + flux(2) *area*dt
         !
       end do
       !$acc end parallel
     end do
     !$acc parallel loop collapse(2) present(my_S_flux,my_N_flux,Stflux,Ntflux) &
     !$acc reduction(+:my_S_flux,my_N_flux)
     do i = my_ips,my_ipe
     do k = my_kps, my_kpe
       my_S_flux = my_S_flux + Stflux(k,i)
       my_N_flux = my_N_flux + Ntflux(k,i)
     end do
     end do

     end select
     !$acc end data
     deallocate(Stflux)
     deallocate(Ntflux)
     !
!     do ibin = 1,nbins
!       call domain_swap_mass_points_2halo_y_GPU (my_c(:,:,:,ibin))
!     end do
     !$acc end data
#else
    !
    !*** Loop over bins
    !
    do ibin = 1, nbins
       !
       select case(CB%time_marching)
          !
       case(TIME_MARCHING_EULER)
          !
          !   Euler O(1)
          !
          do k = my_kps,my_kpe
             do i = my_ips,my_ipe
                ini_c(my_jps_2h:my_jpe_2h,i,k) = my_c (i,my_jps_2h:my_jpe_2h,k,ibin)
                my_c_1d(my_jps_2h:my_jpe_2h) => ini_c (my_jps_2h:my_jpe_2h,i,k)
                my_k_1d(my_jps_2h:my_jpe_2h) => my_k2(my_jps_2h:my_jpe_2h,i,k)
                my_v_1d(my_jbs_1h:my_jbe_1h) => CB%v (my_jbs_1h:my_jbe_1h,i,k)
                !
                ki = KT_RHS(my_c_1d,MY_GRID%dX2_p,MY_GRID%dX2_b,my_v_1d,my_k_1d, &
                     flux,my_jps,my_jpe,my_jbs,my_jbe)
                !
                my_c(i,my_jps:my_jpe,k,ibin) = my_c_1d(my_jps:my_jpe) + dt*ki(my_jps:my_jpe)
                !
                !***     Fluxes
                !
                area = MY_GRID%dX3_b(k)*MY_GRID%dX1_b(i)
                if(my_S_proc.eq.-1) my_S_flux = my_S_flux + flux(1) *area*dt
                if(my_N_proc.eq.-1) my_N_flux = my_N_flux + flux(2) *area*dt
                !
             end do
          end do
          !
       case(TIME_MARCHING_RK)
          !
          !   RK O(4)
          !
          do k = my_kps,my_kpe
             do i = my_ips,my_ipe
                ini_c(my_jps_2h:my_jpe_2h,i,k) = my_c (i,my_jps_2h:my_jpe_2h,k,ibin)
                my_c_1d(my_jps_2h:my_jpe_2h) => ini_c (my_jps_2h:my_jpe_2h,i,k)
                my_k_1d(my_jps_2h:my_jpe_2h) => my_k2(my_jps_2h:my_jpe_2h,i,k)
                my_v_1d(my_jbs_1h:my_jbe_1h) => CB%v (my_jbs_1h:my_jbe_1h,i,k)
                !
                !***     Time integration in the interval (t,t+dt) using a RK4 Runge-Kutta method
                !
                ki = KT_RHS(my_c_1d,MY_GRID%dX2_p,MY_GRID%dX2_b,my_v_1d,my_k_1d, &
                     flux,my_jps,my_jpe,my_jbs,my_jbe)

                ! In this first stage, we need to copy also the boundary halos.
                c0(:,i,k) = my_c_1d(:)
                c0(my_jps:my_jpe,i,k) = c0(my_jps:my_jpe,i,k) + 0.5_rp*dt*ki(my_jps:my_jpe)

                new_c(:,i,k) = my_c_1d(:)
                new_c(my_jps:my_jpe,i,k) = new_c(my_jps:my_jpe,i,k) + dt*ki(my_jps:my_jpe) / 6.0_rp
             end do
          end do
          !
          !      Successive approaches. Note that swap is necessary to update predictor halo values
          !
          call domain_swap_mass_points_2halo_first (c0(:,:,:), my_jps, my_jpe, &
               my_ips,my_ipe,my_kps,my_kpe, my_S_proc, my_N_proc)


          do k = my_kps,my_kpe
             do i = my_ips,my_ipe
                my_c_1d(my_jps_2h:my_jpe_2h) => c0 (my_jps_2h:my_jpe_2h,i,k)
                my_k_1d(my_jps_2h:my_jpe_2h) => my_k2(my_jps_2h:my_jpe_2h,i,k)
                my_v_1d(my_jbs_1h:my_jbe_1h) => CB%v (my_jbs_1h:my_jbe_1h,i,k)
                !
                ki = KT_RHS(my_c_1d,MY_GRID%dX2_p,MY_GRID%dX2_b,my_v_1d,my_k_1d, &
                     flux,my_jps,my_jpe,my_jbs,my_jbe)
                !
                c0(my_jps:my_jpe,i,k) = ini_c(my_jps:my_jpe,i,k) + 0.5_rp*dt*ki(my_jps:my_jpe)
                new_c(my_jps:my_jpe,i,k) = new_c(my_jps:my_jpe,i,k) + dt*ki(my_jps:my_jpe) / 3.0_rp
             end do
          end do
          !
          !      Successive approaches. Note that swap is necessary to update predictor halo values
          !
          call domain_swap_mass_points_2halo_first (c0(:,:,:), my_jps, my_jpe, &
               my_ips,my_ipe,my_kps,my_kpe, my_S_proc, my_N_proc)

          do k = my_kps,my_kpe
             do i = my_ips,my_ipe
                my_c_1d(my_jps_2h:my_jpe_2h) => c0 (my_jps_2h:my_jpe_2h,i,k)
                my_k_1d(my_jps_2h:my_jpe_2h) => my_k2(my_jps_2h:my_jpe_2h,i,k)
                my_v_1d(my_jbs_1h:my_jbe_1h) => CB%v (my_jbs_1h:my_jbe_1h,i,k)

                ki = KT_RHS(my_c_1d,MY_GRID%dX2_p,MY_GRID%dX2_b,my_v_1d,my_k_1d, &
                     flux,my_jps,my_jpe,my_jbs,my_jbe)
                !
                c0(my_jps:my_jpe,i,k) = ini_c(my_jps:my_jpe,i,k) + dt*ki(my_jps:my_jpe)
                new_c(my_jps:my_jpe,i,k) = new_c(my_jps:my_jpe,i,k) + dt*ki(my_jps:my_jpe) / 3.0_rp
             end do
          end do
          !
          !      Successive approaches. Note that swap is necessary to update predictor halo values
          !
          call domain_swap_mass_points_2halo_first (c0(:,:,:), my_jps,my_jpe, &
               my_ips,my_ipe,my_kps,my_kpe, my_S_proc, my_N_proc)

          do k = my_kps,my_kpe
             do i = my_ips,my_ipe
                my_c_1d(my_jps_2h:my_jpe_2h) => c0 (my_jps_2h:my_jpe_2h,i,k)
                my_k_1d(my_jps_2h:my_jpe_2h) => my_k2(my_jps_2h:my_jpe_2h,i,k)
                my_v_1d(my_jbs_1h:my_jbe_1h) => CB%v (my_jbs_1h:my_jbe_1h,i,k)

                ki = KT_RHS(my_c_1d,MY_GRID%dX2_p,MY_GRID%dX2_b,my_v_1d,my_k_1d, &
                     flux,my_jps,my_jpe,my_jbs,my_jbe)
                !
                my_c(i,my_jps:my_jpe,k,ibin) = new_c(my_jps:my_jpe,i,k) + dt*ki(my_jps:my_jpe) / 6.0_rp
                !
                !***     Fluxes
                !
                area = MY_GRID%dX3_b(k)*MY_GRID%dX1_b(i)
                if(my_S_proc.eq.-1) my_S_flux = my_S_flux + flux(1) *area*dt
                if(my_N_proc.eq.-1) my_N_flux = my_N_flux + flux(2) *area*dt
                !
             end do
          end do
          !
       end select
       !
    end do
    !
    deallocate(ki )
!    do ibin = 1,nbins
!      call domain_swap_mass_points_2halo_y (my_c(:,:,:,ibin))
!    end do
    !
#endif

     return
     end subroutine ADS_solve_along_y
!
!
!-----------------------------------
!    subroutine ADS_solve_along_z
!-----------------------------------
!
!>   @details
!>   Solves a time step dt of the 1D ADS equation along z direction.
!
#if defined WITH_ACC
     subroutine ADS_solve_along_z(my_D_flux,my_U_flux,dt,my_c,my_acum,my_w1,my_w2,stime,my_vs,nbins)
#else
  subroutine ADS_solve_along_z(my_D_flux,my_U_flux,dt,my_c,my_acum,my_k3,my_w1,my_w2,stime,my_vs,MY_GRID,nbins)
#endif
     implicit none
!
!>   @param my_D_flux increment of mass flux across the D domain boundary during dt
!>   @param my_U_flux increment of mass flux across the U domain boundary during dt
!>   @param dt        time increment
!>   @param my_c      my values of scaled transported variable (concentration) at mass points
!>   @param my_acum   my values of ground accumulation at mass points
!>   @param my_k3     my values of scaled diffusion at mass points
!>   @param my_w1     my values of scaled w-velocity at boundaries, first snapshot
!>   @param my_w2     my values of scaled w-velocity at boundaries, second snapshot
!>   @param stime     position between snapshots
!>   @param my_vs     settling velocity at w-boundaries
    !>   @param MY_GRID   grid structure
!>   @param nbins     number of bins
!
     integer(ip),         intent(IN   ) :: nbins
     real(rp),            intent(INOUT) :: my_D_flux
     real(rp),            intent(INOUT) :: my_U_flux
     real(rp),            intent(IN   ) :: dt
     real(rp),target     ,intent(INOUT) :: my_c   (my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h,my_kps_2h:my_kpe_2h,nbins)
     real(rp),            intent(INOUT) :: my_acum(my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h,nbins)
     real(rp),            intent(IN   ) :: my_w1  (my_kbs_1h:my_kbe_1h,my_ips   :my_ipe   ,my_jps   :my_jpe)
     real(rp),            intent(IN   ) :: my_w2  (my_kbs_1h:my_kbe_1h,my_ips   :my_ipe   ,my_jps   :my_jpe)
     real(rp),            intent(IN   ) :: stime
     real(rp),            intent(IN   ) :: my_vs  (my_ips   :my_ipe   ,my_jps   :my_jpe   ,my_kbs_1h:my_kbe_1h,nbins)
#ifndef WITH_ACC
     type(ARAKAWA_C_GRID),intent(IN   ) :: MY_GRID
     real(rp),target     ,intent(IN   ) :: my_k3  (my_kps_2h:my_kpe_2h,my_ips   :my_ipe   ,my_jps   :my_jpe)
#endif
#if defined WITH_ACC
     integer(ip)           :: limtype,k,idx
     real(rp),allocatable  :: Dtflux(:,:),Utflux(:,:)
#else
    real(rp), pointer     :: my_c_1d(:)
    real(rp), pointer     :: my_w_1d(:)
    real(rp), pointer     :: my_k_1d(:)
    real(rp), pointer     :: c0(:,:,:)
    real(rp), pointer     :: new_c(:,:,:)
    real(rp), pointer     :: ini_c(:,:,:)
    real(rp), allocatable :: ki (:)
#endif
     integer(ip)           :: i,j,ibin
     real(rp)              :: flux(2),area

     !$acc data &
     !$acc present(my_vs) &
     !$acc present(my_w1,my_w2)  &
     !$acc present(zinic,znewc,zc0,dX3_b,dX3_p,ADS_w,ADS_k3,dX2_b,dX1_b,Dswapi,Dswapo,Uswapi,Uswapo) &
     !$acc present(my_acum) &
     !$acc present(my_c(:,:,:,:)) &
     !$acc present(my_D_flux,my_U_flux) &
     !$acc present(CB,CB%bvalue(:))

#if defined WITH_ACC
     limtype = CB%limitertype

     allocate(Dtflux(my_ips:my_ipe,my_jps:my_jpe))
     allocate(Utflux(my_ips:my_ipe,my_jps:my_jpe))
     !$acc data create(Dtflux,Utflux)
     !$acc kernels default(none)
     Dtflux(:,:) = 0.0_rp
     Utflux(:,:) = 0.0_rp
     !$acc end kernels
#else
    !
    !*** Allocate memory
    !
    allocate(ki     (my_kps:my_kpe))
    !
    !*** Prepare shape of temporary fields
    !
    c0(my_kps_2h:my_kpe_2h,my_ips_2h:my_ipe_2h, my_jps_2h:my_jpe_2h) => CB%c0(:)
    new_c(my_kps_2h:my_kpe_2h,my_ips_2h:my_ipe_2h, my_jps_2h:my_jpe_2h) => CB%new_c(:)
    ini_c(my_kps_2h:my_kpe_2h,my_ips_2h:my_ipe_2h, my_jps_2h:my_jpe_2h) => CB%raw_c(:)
    !
#endif
! The main difference between CPU & GPU versions is that the "Loop over bins" is
! done outside (outer one) the select case in the former, whereas in the latter such loop is inserted inside the cases
#if defined WITH_ACC
!
!*** Loop over bins
!
     do ibin = 1,nbins
       !
       call domain_swap_mass_points_2halo_z_GPU (my_c(:,:,:,ibin))
!
!***   Compute current velocity component (required for applying boundary conditions)
       !$acc kernels default(none)
       !$acc loop independent
       do j=my_jps,my_jpe
       do i=my_ips,my_ipe
         ADS_w(:,i,j,ibin) = (1.0_rp-stime) * my_w1(:,i,j) + stime * my_w2(:,i,j)
         ADS_w(:,i,j,ibin) = ADS_w(:,i,j,ibin) - my_vs(i,j,:,ibin)
       end do
       end do
       !$acc end kernels
!
!***   Update current w component in order to be used by boundary conditions
       CB%iw(my_kbs_1h:my_kbe_1h,my_ips:my_ipe,my_jps:my_jpe)=>ADS_w(:,:,:,ibin)
!
!***  Apply boundaries
!
       call CB%apply_bconditions( my_c(:,:,:,ibin), 3_ip )
     end do
       !
     select case(CB%time_marching)
     !
     case(TIME_MARCHING_EULER)
       !
       !   Euler O(1)
       !
     do ibin = 1,nbins

       !$acc parallel loop independent gang worker num_workers(16) &
       !$acc private(rk,zki,flux,j,i,k,area)
       do idx = 1, (my_ipe-my_ips+1)*(my_jpe-my_jps+1)
         i = my_ips + mod(idx-1,my_ipe-my_ips+1)
         j = my_jps + (idx-1) / (my_ipe-my_ips+1)

         !$acc loop vector
         do k = my_kps_2h,my_kpe_2h
           zinic(k,i,j) = my_c(i,j,k,ibin)
         end do

         call KT_RHS(zki,zinic (my_kps_2h:my_kpe_2h,i,j), &
                     dX3_p,dX3_b, &
                     ADS_w(my_kbs_1h:my_kbe_1h,i,j,ibin), &
                     ADS_k3(my_kps_2h:my_kpe_2h,i,j), &
                     flux,my_kps,my_kpe,my_kbs,my_kbe, &
                     rk, limtype)

         !$acc loop vector
         do k = my_kps,my_kpe
           my_c(i,j,k,ibin) = my_c(i,j,k,ibin) + dt*zki(k)
         end do
!
!***   Fluxes
!
         area = dX2_b(j)*dX1_b(i) !/(MY_GRID%Hm1_p(j)*MY_GRID%Hm2_p(j))
         if(my_D_proc.eq.-1) then
            Dtflux(i,j)    = flux(1)*area*dt              ! total mass
            my_acum(i,j,ibin) = my_acum(i,j,ibin) + abs(flux(1))*dt ! mass per unit ara (force positive sign)
         end if
         if(my_U_proc.eq.-1) Utflux(i,j) = Utflux(i,j) + flux(2) *area*dt
         !
       end do
       !$acc end parallel
     end do
     !$acc parallel loop collapse(2) present(my_D_flux,my_U_flux,Dtflux,Utflux) &
     !$acc reduction(+:my_D_flux,my_U_flux)
     do j = my_jps,my_jpe
     do i = my_ips,my_ipe
       my_D_flux = my_D_flux + Dtflux(i,j)
       my_U_flux = my_U_flux + Utflux(i,j)
     end do
     end do

       !
     case(TIME_MARCHING_RK)
       !
       !   RK O(4)
       !

     do ibin = 1,nbins

       !$acc parallel loop gang worker num_workers(16) private(rk,zki,flux,j,i)
       do idx = 1, (my_ipe-my_ips+1)*(my_jpe-my_jps+1)
         i = my_ips + mod(idx-1,my_ipe-my_ips+1)
         j = my_jps + (idx-1) / (my_ipe-my_ips+1)

         ! In this first stage, we need to copy also the boundary halos
         ! in new_c and c0.
         !$acc loop vector
         do k = my_kps_2h,my_kpe_2h
           zinic(k,i,j) = my_c(i,j,k,ibin)
           znewc(k,i,j) = my_c(i,j,k,ibin)
           zc0(k,i,j) = my_c(i,j,k,ibin)
         end do
!
!***     Time integration in the interval (t,t+dt) using a RK4 Runge-Kutta method
!
         call KT_RHS(zki,zinic (my_kps_2h:my_kpe_2h,i,j), &
                     dX3_p,dX3_b, &
                     ADS_w(my_kbs_1h:my_kbe_1h,i,j,ibin), &
                     ADS_k3(my_kps_2h:my_kpe_2h,i,j), &
                     flux,my_kps,my_kpe,my_kbs,my_kbe, &
                     rk, limtype)

         !$acc loop vector
         do k = my_kps,my_kpe
           zc0(k,i,j) = zc0(k,i,j) + 0.5_rp*dt*zki(k)

           znewc(k,i,j) = znewc(k,i,j) + dt*zki(k) / 6.0_rp
         end do

         if (my_D_proc.ne.-1) Dswapo(1:2,i,j) = zc0(my_kps:my_kps+1,i,j)
         if (my_U_proc.ne.-1) Uswapo(1:2,i,j) = zc0(my_kpe-1:my_kpe,i,j)
       end do
       !$acc end parallel
!
!      Successive approaches. Note that swap is necessary to update predictor halo values
!
       call domain_swap_mass_points_2halo_first_GPU (Dswapi, Dswapo, Uswapi,Uswapo, &
           my_ips,my_ipe,my_jps,my_jpe, my_D_proc, my_U_proc)

       !$acc parallel loop independent gang worker num_workers(16) private(rk,zki,flux,j,i)
       do idx = 1, (my_ipe-my_ips+1)*(my_jpe-my_jps+1)
         i = my_ips + mod(idx-1,my_ipe-my_ips+1)
         j = my_jps + (idx-1) / (my_ipe-my_ips+1)

         if (my_D_proc.ne.-1) zc0(my_kps-2:my_kps-1,i,j) = Dswapi(1:2,i,j)
         if (my_U_proc.ne.-1) zc0(my_kpe+1:my_kpe+2,i,j) = Uswapi(1:2,i,j)

         call KT_RHS(zki,zc0(my_kps_2h:my_kpe_2h,i,j), &
                     dX3_p,dX3_b, &
                     ADS_w(my_kbs_1h:my_kbe_1h,i,j,ibin), &
                     ADS_k3(my_kps_2h:my_kpe_2h,i,j), &
                     flux,my_kps,my_kpe,my_kbs,my_kbe, &
                     rk, limtype)

         !$acc loop independent vector
         do k = my_kps,my_kpe
           zc0(k,i,j) = zinic(k,i,j) + 0.5_rp*dt*zki(k)
           znewc(k,i,j) = znewc(k,i,j) + dt*zki(k) / 3.0_rp
         end do

         if (my_D_proc.ne.-1) Dswapo(1:2,i,j) = zc0(my_kps:my_kps+1,i,j)
         if (my_U_proc.ne.-1) Uswapo(1:2,i,j) = zc0(my_kpe-1:my_kpe,i,j)
       end do
       !$acc end parallel
!
!      Successive approaches. Note that swap is necessary to update predictor halo values
!
       call domain_swap_mass_points_2halo_first_GPU (Dswapi, Dswapo, Uswapi,Uswapo, &
           my_ips,my_ipe,my_jps,my_jpe, my_D_proc, my_U_proc)

       !$acc parallel loop independent gang worker num_workers(16) &
       !$acc private(rk,zki,flux,j,i)
       do idx = 1, (my_ipe-my_ips+1)*(my_jpe-my_jps+1)
         i = my_ips + mod(idx-1,my_ipe-my_ips+1)
         j = my_jps + (idx-1) / (my_ipe-my_ips+1)

         if (my_D_proc.ne.-1) zc0(my_kps-2:my_kps-1,i,j) = Dswapi(1:2,i,j)
         if (my_U_proc.ne.-1) zc0(my_kpe+1:my_kpe+2,i,j) = Uswapi(1:2,i,j)

         call KT_RHS(zki, zc0(my_kps_2h:my_kpe_2h,i,j), &
                     dX3_p,dX3_b, &
                     ADS_w(my_kbs_1h:my_kbe_1h,i,j,ibin), &
                     ADS_k3(my_kps_2h:my_kpe_2h,i,j), &
                     flux,my_kps,my_kpe,my_kbs,my_kbe, &
                     rk, limtype )

         !$acc loop independent vector
         do k = my_kps,my_kpe
           zc0(k,i,j) = zinic(k,i,j) + dt*zki(k)
           znewc(k,i,j) = znewc(k,i,j) + dt*zki(k) / 3.0_rp
         end do

         if (my_D_proc.ne.-1) Dswapo(1:2,i,j) = zc0(my_kps:my_kps+1,i,j)
         if (my_U_proc.ne.-1) Uswapo(1:2,i,j) = zc0(my_kpe-1:my_kpe,i,j)
       end do
       !$acc end parallel
!
!      Successive approaches. Note that swap is necessary to update predictor halo values
!
       call domain_swap_mass_points_2halo_first_GPU (Dswapi, Dswapo, Uswapi,Uswapo, &
           my_ips,my_ipe,my_jps,my_jpe, my_D_proc, my_U_proc)

       !$acc parallel loop independent gang worker num_workers(16) &
       !$acc private(rk,zki,flux,j,i,k,area)
       do idx = 1, (my_ipe-my_ips+1)*(my_jpe-my_jps+1)
         i = my_ips + mod(idx-1,my_ipe-my_ips+1)
         j = my_jps + (idx-1) / (my_ipe-my_ips+1)

         if (my_D_proc.ne.-1) zc0(my_kps-2:my_kps-1,i,j) = Dswapi(1:2,i,j)
         if (my_U_proc.ne.-1) zc0(my_kpe+1:my_kpe+2,i,j) = Uswapi(1:2,i,j)

         call KT_RHS(zki, zc0(my_kps_2h:my_kpe_2h,i,j), &
                     dX3_p,dX3_b, &
                     ADS_w(my_kbs_1h:my_kbe_1h,i,j,ibin), &
                     ADS_k3(my_kps_2h:my_kpe_2h,i,j), &
                     flux,my_kps,my_kpe,my_kbs,my_kbe, &
                     rk, limtype )
!
         !$acc loop vector
         do k = my_kps,my_kpe
           my_c(i,j,k,ibin) = znewc(k,i,j) + dt*zki(k) / 6.0_rp
         end do
!
!***   Fluxes
!
         area = dX2_b(j)*dX1_b(i) !/(MY_GRID%Hm1_p(j)*MY_GRID%Hm2_p(j))
         if(my_D_proc.eq.-1) then
            Dtflux(i,j)    = Dtflux(i,j) + flux(1)*area*dt              ! total mass
            my_acum(i,j,ibin) = my_acum(i,j,ibin) + abs(flux(1))*dt ! mass per unit ara (force positive sign)
         end if
         if(my_U_proc.eq.-1) Utflux(i,j) = Utflux(i,j) + flux(2) *area*dt
         !
       end do
       !$acc end parallel
     end do
     !$acc parallel loop collapse(2) present(my_D_flux,my_U_flux,Dtflux,Utflux) reduction(+:my_D_flux,my_U_flux)
     do j = my_jps,my_jpe
     do i = my_ips,my_ipe
       my_D_flux = my_D_flux + Dtflux(i,j)
       my_U_flux = my_U_flux + Utflux(i,j)
     end do
     end do
      !
      !
     end select
     !$acc end data

     deallocate(Dtflux)
     deallocate(Utflux)
     !
!     do ibin = 1,nbins
!       call domain_swap_mass_points_2halo_z_GPU (my_c(:,:,:,ibin))
!     end do
     !$acc end data
#else
    !
    !*** Loop over bins
    !
    do ibin = 1,nbins
       !
       call domain_swap_mass_points_2halo_z (my_c(:,:,:,ibin))
       !
       !***   Compute current velocity component
       !
       CB%w(:,:,:) = (1.0_rp-stime) * my_w1(:,:,:) + stime * my_w2(:,:,:)
       do j=my_jps,my_jpe
          do i=my_ips,my_ipe
             CB%w(:,i,j) = CB%w(:,i,j) - my_vs(i,j,:,ibin)
          end do
       end do
       !
       !***  Apply boundaries
       !
       call CB%apply_bconditions( my_c(:,:,:,ibin), 3_ip )
       !
       select case(CB%time_marching)
          !
       case(TIME_MARCHING_EULER)
          !
          !   Euler O(1)
          !
          do j = my_jps,my_jpe
             do i = my_ips,my_ipe
                ini_c(my_kps_2h:my_kpe_2h,i,j) = my_c (i,j,my_kps_2h:my_kpe_2h,ibin)
                my_c_1d(my_kps_2h:my_kpe_2h) => ini_c (my_kps_2h:my_kpe_2h,i,j)
                my_k_1d(my_kps_2h:my_kpe_2h) => my_k3(my_kps_2h:my_kpe_2h,i,j)
                my_w_1d(my_kbs_1h:my_kbe_1h) => CB%w(my_kbs_1h:my_kbe_1h,i,j)
                !
                ki = KT_RHS(my_c_1d,MY_GRID%dX3_p,MY_GRID%dX3_b,my_w_1d,my_k_1d, &
                     flux,my_kps,my_kpe,my_kbs,my_kbe)

                my_c(i,j,my_kps:my_kpe,ibin) = my_c_1d(my_kps:my_kpe) + dt*ki(my_kps:my_kpe)
                !
                !***   Fluxes
                !
                area = MY_GRID%dX2_b(j)*MY_GRID%dX1_b(i) !/(MY_GRID%Hm1_p(j)*MY_GRID%Hm2_p(j))
                if(my_D_proc.eq.-1) then
                   my_D_flux    = my_D_flux + flux(1)*area*dt              ! total mass
                   my_acum(i,j,ibin) = my_acum(i,j,ibin) + abs(flux(1))*dt ! mass per unit ara (force positive sign)
                end if
                if(my_U_proc.eq.-1) then
                   my_U_flux = my_U_flux + flux(2) *area*dt
                end if
                !
             end do
          end do
          !
       case(TIME_MARCHING_RK)
          !
          !   RK O(4)
          !
          do j = my_jps,my_jpe
             do i = my_ips,my_ipe
                ini_c(my_kps_2h:my_kpe_2h,i,j) = my_c (i,j,my_kps_2h:my_kpe_2h,ibin)
                my_c_1d(my_kps_2h:my_kpe_2h) => ini_c (my_kps_2h:my_kpe_2h,i,j)
                my_k_1d(my_kps_2h:my_kpe_2h) => my_k3(my_kps_2h:my_kpe_2h,i,j)
                my_w_1d(my_kbs_1h:my_kbe_1h) => CB%w(my_kbs_1h:my_kbe_1h,i,j)
                !
                !***     Time integration in the interval (t,t+dt) using a RK4 Runge-Kutta method
                !
                ki = KT_RHS(my_c_1d,MY_GRID%dX3_p,MY_GRID%dX3_b,my_w_1d,my_k_1d, &
                     flux,my_kps,my_kpe,my_kbs,my_kbe)

                ! In this first stage, we need to copy also the boundary halos.
                c0(:,i,j) = my_c_1d(:)
                c0(my_kps:my_kpe,i,j) = c0(my_kps:my_kpe,i,j) + 0.5_rp*dt*ki(my_kps:my_kpe)

                new_c(:,i,j) = my_c_1d(:)
                new_c(my_kps:my_kpe,i,j) = new_c(my_kps:my_kpe,i,j) + dt*ki(my_kps:my_kpe) / 6.0_rp
             end do
          end do
          !
          !      Successive approaches. Note that swap is necessary to update predictor halo values
          !
          call domain_swap_mass_points_2halo_first (c0(:,:,:), my_kps,my_kpe, &
               my_ips,my_ipe,my_jps,my_jpe, my_D_proc, my_U_proc)

          do j = my_jps,my_jpe
             do i = my_ips,my_ipe
                my_c_1d(my_kps_2h:my_kpe_2h) => c0 (my_kps_2h:my_kpe_2h,i,j)
                my_k_1d(my_kps_2h:my_kpe_2h) => my_k3(my_kps_2h:my_kpe_2h,i,j)
                my_w_1d(my_kbs_1h:my_kbe_1h) => CB%w(my_kbs_1h:my_kbe_1h,i,j)

                ki = KT_RHS(my_c_1d,MY_GRID%dX3_p,MY_GRID%dX3_b,my_w_1d,my_k_1d, &
                     flux,my_kps,my_kpe,my_kbs,my_kbe)
                !
                !
                c0(my_kps:my_kpe,i,j) = ini_c(my_kps:my_kpe,i,j) + 0.5_rp*dt*ki(my_kps:my_kpe)
                new_c(my_kps:my_kpe,i,j) = new_c(my_kps:my_kpe,i,j) + dt*ki(my_kps:my_kpe) / 3.0_rp
             end do
          end do
          !
          call domain_swap_mass_points_2halo_first (c0(:,:,:), my_kps, my_kpe, &
               my_ips,my_ipe,my_jps,my_jpe, my_D_proc, my_U_proc)

          do j = my_jps,my_jpe
             do i = my_ips,my_ipe
                my_c_1d(my_kps_2h:my_kpe_2h) => c0 (my_kps_2h:my_kpe_2h,i,j)
                my_k_1d(my_kps_2h:my_kpe_2h) => my_k3(my_kps_2h:my_kpe_2h,i,j)
                my_w_1d(my_kbs_1h:my_kbe_1h) => CB%w(my_kbs_1h:my_kbe_1h,i,j)

                ki = KT_RHS(my_c_1d,MY_GRID%dX3_p,MY_GRID%dX3_b,my_w_1d,my_k_1d, &
                     flux,my_kps,my_kpe,my_kbs,my_kbe)
                !
                c0(my_kps:my_kpe,i,j) = ini_c(my_kps:my_kpe,i,j) + dt*ki(my_kps:my_kpe)
                new_c(my_kps:my_kpe,i,j) = new_c(my_kps:my_kpe,i,j) + dt*ki(my_kps:my_kpe) / 3.0_rp
             end do
          end do
          !
          call domain_swap_mass_points_2halo_first (c0(:,:,:), my_kps,my_kpe, &
               my_ips,my_ipe,my_jps,my_jpe, my_D_proc, my_U_proc)

          do j = my_jps,my_jpe
             do i = my_ips,my_ipe
                my_c_1d(my_kps_2h:my_kpe_2h) => c0 (my_kps_2h:my_kpe_2h,i,j)
                my_k_1d(my_kps_2h:my_kpe_2h) => my_k3(my_kps_2h:my_kpe_2h,i,j)
                my_w_1d(my_kbs_1h:my_kbe_1h) => CB%w(my_kbs_1h:my_kbe_1h,i,j)

                ki = KT_RHS(my_c_1d,MY_GRID%dX3_p,MY_GRID%dX3_b,my_w_1d,my_k_1d, &
                     flux,my_kps,my_kpe,my_kbs,my_kbe)
                !
                my_c(i,j,my_kps:my_kpe,ibin) = new_c(my_kps:my_kpe,i,j) + dt*ki(my_kps:my_kpe) / 6.0_rp
                !
                !***   Fluxes
                !
                area = MY_GRID%dX2_b(j)*MY_GRID%dX1_b(i) !/(MY_GRID%Hm1_p(j)*MY_GRID%Hm2_p(j))
                if(my_D_proc.eq.-1) then
                   my_D_flux    = my_D_flux + flux(1)*area*dt              ! total mass
                   my_acum(i,j,ibin) = my_acum(i,j,ibin) + abs(flux(1))*dt ! mass per unit ara (force positive sign)
                end if
                if(my_U_proc.eq.-1) then
                   my_U_flux = my_U_flux + flux(2) *area*dt
                end if
                !
             end do
          end do
          !
       end select
       !
    end do
    !
    deallocate(ki )
    !
!    do ibin = 1,nbins
!      call domain_swap_mass_points_2halo_z (my_c(:,:,:,ibin))
!    end do
#endif
     !
     return
     end subroutine ADS_solve_along_z
!
! The main difference between CPU & GPU versions is tha KT_RHS is a function in
! the former, whereas in the latter KT_RHS is a subroutine
#if defined WITH_ACC
     subroutine KT_RHS(ki,c,dxp,dxb,u,k,flux,ips,ipe,ibs,ibe,r,limitertype)
#else
  function KT_RHS(c,dxp,dxb,u,k,flux,ips,ipe,ibs,ibe)
#endif
!******************************************************************
!*
!*   Computes the RHS for the Kuganov-Tadmor (KT) scheme at mass points
!*   Note that the RHS at ghost mass poins is not computed
!*
!******************************************************************
     !$acc routine vector
     implicit none
     integer(ip), intent(in)    :: ips,ipe,ibs,ibe
     real   (rp), intent(in)    :: c      (ips-2:ipe+2)
     real   (rp), intent(in)    :: dxp    (ips-2:ipe+2)   ! dxp(i) distance between point i and i+1
     real   (rp), intent(in)    :: dxb    (ibs-1:ibe+1)   ! dxb(i) size of the cell       i
     real   (rp), intent(in)    :: u      (ibs-1:ibe+1)   ! velocity at boundaries
     real   (rp), intent(in)    :: k      (ips-2:ipe+2)   ! diffusion at mass points
#if defined WITH_ACC
     real    (rp), intent(out)  :: ki(ips:ipe)
     real    (rp)               :: r      (ips-2:ipe+2)
     integer (ip), intent(in)   :: limitertype
#endif
!
!    work arrays
!
#if defined WITH_ACC
    real   (rp), intent(out) :: flux   (2          )   ! Flux at boundaries
    real   (rp)          :: c_r,c_l,c_r2,c_l2
#else
    real   (rp), intent(inout) :: flux   (2          )   ! Flux at boundaries
    real   (rp)                :: KT_RHS (ips  :ipe  )
    real   (rp), pointer :: r(:)
    real   (rp), pointer :: c_r(:)
    real   (rp), pointer :: c_l(:)
#endif
     integer(ip)          :: i
     real   (rp)          :: F_ml,F_mr,F_pl,F_pr,F_m,F_p,P_m,P_p

#ifndef WITH_ACC
    r(ips-2:ipe+2) => CB%raw_r(1:ipe-ips+5)
    c_r(ips-1:ipe+1) => CB%raw_c_r(1:ipe-ips+3)
    c_l(ips-1:ipe+1) => CB%raw_c_l(1:ipe-ips+3)
    !
    !*** Computes flux limiter function
    !
    r = CB%flux_limiter(c,ips,ipe)
    !
!!
!    !*** Computes c_r (right) and cl (left) at boundaries
!    !*** (i.e. at i+1/2 and i-1/2 for each cell)
!    !
    do i = ips-1,ipe+1
       c_r(i) = c(i  ) - 0.5_rp*r(i  )*(c(i+1)-c(i  ))
       c_l(i) = c(i-1) + 0.5_rp*r(i-1)*(c(i  )-c(i-1))
    end do
#endif
!
!*** Computes flux limiter function
!
!     select case(limitertype)
!     case(1)
!        !                         minmod
!        r = r_minmod(c,ips,ipe)
!     case(2)
        !                         superbee
#if defined WITH_ACC
        call r_sbee2(c,ips,ipe,r)
#endif
!     case(3)
!        !                         ospre
!        r = r_ospre(c,ips,ipe)
!     end select
!
!*** Computes c_r (right) and cl (left) at boundaries
!*** (i.e. at i+1/2 and i-1/2 for each cell)
!
!     do i = ips-1,ipe+1
!        c_r(i) = c(i  ) - 0.5_rp*r(i  )*(c(i+1)-c(i  ))
!        c_l(i) = c(i-1) + 0.5_rp*r(i-1)*(c(i  )-c(i-1))
!     end do
!
!*** Computes RHS at mass points
!
     !$acc loop independent vector private(c_r,c_l,c_r2,c_l2,F_ml,F_mr,F_pl,F_pr,F_m,F_p,P_m,P_p)
     do i = ips,ipe
#if defined WITH_ACC
        c_r = c(i  ) - 0.5_rp*r(i  )*(c(i+1)-c(i  ))
        c_l = c(i-1) + 0.5_rp*r(i-1)*(c(i  )-c(i-1))
        c_r2 = c(i+1) - 0.5_rp*r(i+1 )*(c(i+2)-c(i+1 ))
        c_l2 = c(i) + 0.5_rp*r(i)*(c(i+1)-c(i))
        !
        F_ml = c_l*u(i  )   ! left  advecive flux at i-1/2
        F_mr = c_r*u(i  )   ! right advecive flux at i-1/2
        F_pl = c_l2*u(i+1)   ! left  advecive flux at i+1/2
        F_pr = c_r2*u(i+1)   ! right advecive flux at i+1/2
        F_m = 0.5_rp*(F_ml+F_mr) - 0.5_rp*abs(u(i  ))*(c_r-c_l)    ! F at i-1/2
        F_p = 0.5_rp*(F_pl+F_pr) - 0.5_rp*abs(u(i+1))*(c_r2-c_l2)    ! F at i+1/2
#else
        F_ml = c_l(i  )*u(i  )   ! left  advecive flux at i-1/2
        F_mr = c_r(i  )*u(i  )   ! right advecive flux at i-1/2
        F_pl = c_l(i+1)*u(i+1)   ! left  advecive flux at i+1/2
        F_pr = c_r(i+1)*u(i+1)   ! right advecive flux at i+1/2
        F_m = 0.5_rp*(F_ml+F_mr) - 0.5_rp*abs(u(i  ))*(c_r(i  )-c_l(i  ))    ! F at i-1/2
        F_p = 0.5_rp*(F_pl+F_pr) - 0.5_rp*abs(u(i+1))*(c_r(i+1)-c_l(i+1))    ! F at i+1/2
#endif

        !
        P_m = 0.5_rp*(k(i  )+k(i-1))*(c(i  )-c(i-1))/dxp(i-1)                ! P at i-1/2
        P_p = 0.5_rp*(k(i+1)+k(i  ))*(c(i+1)-c(i  ))/dxp(i  )                ! P at i+1/2

#if defined WITH_ACC
        ki(i) = -1.0_rp*(F_p-F_m)/dxb(i) + (P_p-P_m)/dxb(i)
#else
        KT_RHS(i) = -1.0_rp*(F_p-F_m)/dxb(i) + (P_p-P_m)/dxb(i)
#endif

        if(i.eq.ips) flux(1)=F_m+P_m
        if(i.eq.ipe) flux(2)=F_p+P_p

     end do

#if defined WITH_ACC
     end subroutine KT_RHS
#else
     end function KT_RHS
#endif
!
!
!
     function r_minmod(c,ips,ipe)
!******************************************************************
!*
!*   Returns the minmod limiter function at all mass points
!*   (including the 2 ghost poins)
!*
!******************************************************************
     !$acc routine seq
     implicit none
     integer(ip), intent(in)    :: ips,ipe
     real   (rp), intent(in)    :: c       (ips-2:ipe+2)
     real   (rp)                :: r_minmod(ips-2:ipe+2)
!
     integer(ip) :: i
     real   (rp) :: r
!
     do i = ips-1,ipe+1
        if(abs(c(i+1)-c(i)).lt.epsilon) then
           r_minmod(i) = 1.0_rp
        else
           r           = (c(i)-c(i-1))/(c(i+1)-c(i))
   !        r           = r *dxp(i)/dxp(i-1)   ! non uniform
           r_minmod(i) = max(0.0_rp,min(1.0_rp,r))
        end if
     end do
     r_minmod(ips-2) = r_minmod(ips-1)
     r_minmod(ipe+2) = r_minmod(ipe+1)
!
     end function r_minmod
!
!
!
     function r_sbee(c,ips,ipe)
!******************************************************************
!*
!*   Returns the superbee flux limiter function at all mass points
!*   (including the 2 ghost poins)
!*
!******************************************************************
     !$acc routine seq
     implicit none
     integer(ip), intent(in)    :: ips,ipe
     real   (rp), intent(in)    :: c     (ips-2:ipe+2)
     real   (rp)                :: r_sbee(ips-2:ipe+2)
!
     integer(ip) :: i
     real   (rp) :: r
!
     do i = ips-1,ipe+1
        if(abs(c(i+1)-c(i)).lt.epsilon) then
           if(abs(c(i)-c(i-1)).lt.epsilon) then
             r_sbee(i) = 0.0_rp
           else
             r_sbee(i) = 2.0_rp
           end if
        else
           r         = (c(i)-c(i-1))/(c(i+1)-c(i))
 !          r         = r *dxp(i)/dxp(i-1)   ! non uniform

           r_sbee(i) = max(0.0_rp,min(2.0_rp*r,1.0_rp),min(r,2.0_rp))
        end if
     end do
     r_sbee(ips-2) = r_sbee(ips-1)
     r_sbee(ipe+2) = r_sbee(ipe+1)
!
     end function r_sbee
!
!
!
     subroutine r_sbee2(c,ips,ipe,sbee)
!******************************************************************
!*
!*   Returns the superbee flux limiter function at all mass points
!*   (including the 2 ghost poins)
!*
!******************************************************************
     !$acc routine vector
     implicit none
     integer(ip), intent(in)    :: ips,ipe
     real   (rp), intent(in)    :: c     (ips-2:ipe+2)
     !real   (rp)                :: r_sbee2(ips-2:ipe+2)
     real   (rp), intent(out)   :: sbee(ips-2:ipe+2)
!
     integer(ip) :: e
     real   (rp) :: r
!
     !$acc loop independent vector private(r)
     do e = ips-1,ipe+1
        if(abs(c(e+1)-c(e)).lt.epsilon) then
           if(abs(c(e)-c(e-1)).lt.epsilon) then
             sbee(e) = 0.0_rp
           else
             sbee(e) = 2.0_rp
           end if
        else
           r         = (c(e)-c(e-1))/(c(e+1)-c(e))
 !          r         = r *dxp(e)/dxp(e-1)   ! non uniform

           sbee(e) = max(0.0_rp,min(2.0_rp*r,1.0_rp),min(r,2.0_rp))
        end if
     end do
     sbee(ips-2) = sbee(ips-1)
     sbee(ipe+2) = sbee(ipe+1)
!
     end subroutine r_sbee2
!
!
!
     function r_ospre(c,ips,ipe)
!******************************************************************
!*
!*   Returns the ospre flux limiter function at all mass points
!*   (including the 2 ghost poins)
!*
!******************************************************************
     !$acc routine seq
     implicit none
     integer(ip), intent(in)    :: ips,ipe
     real   (rp), intent(in)    :: c      (ips-2:ipe+2)
     real   (rp)                :: r_ospre(ips-2:ipe+2)
!
     integer(ip) :: i
     real   (rp) :: r
!
     do i = ips-1,ipe+1
        if(abs(c(i+1)-c(i)).lt.epsilon) then
           if(abs(c(i)-c(i-1)).lt.epsilon) then
             r_ospre(i) = 0.0_rp
           else
             r_ospre(i) = 1.5_rp
           end if
        else
           r          = (c(i)-c(i-1))/(c(i+1)-c(i))
   !       r          = r *dxp(i)/dxp(i-1)   ! non uniform

           r_ospre(i) = 1.5_rp*(r*r+r)/(r*r+r+1.0_rp)
        end if
     end do
     r_ospre(ips-2) = r_ospre(ips-1)
     r_ospre(ipe+2) = r_ospre(ipe+1)
!
     end function r_ospre
!
!
  subroutine freeflow( c, mydim )
!******************************************************************
!*
!*   Apply free flow boundary conditions
!*
!******************************************************************
     implicit none
     real   (rp), intent(inout) :: c  (my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h,my_kps_2h:my_kpe_2h)
     integer(ip), intent(in)    :: mydim

     integer(ip)                :: i,j,k
     ! In X dimension
       !$acc data present(ADS_u(:,:,:),ADS_v(:,:,:)) &
       !$acc present(CB) &
       !$acc present(CB%bvalue(:)) &
       !$acc present(c(:,:,:))
     if (mydim.eq.1) then
       if (my_W_proc.eq.-1) then
               !$acc parallel loop gang vector collapse(2)  default(none)
         do k = my_kps,my_kpe
         do j = my_jps,my_jpe
#if defined WITH_ACC
           if (ADS_u(my_ibs,j,k).gt.0.0_rp) then         ! inflow (dirichlet)
#else
           if (CB%u(my_ibs,j,k).gt.0.0_rp) then         ! inflow (dirichlet)
#endif
             c(my_ips-1,j,k) = 2.0_rp*CB%bvalue(1)-c(my_ips,j,k)
           else                                         ! outflow (newman, zero gradient)
             c(my_ips-1,j,k) = c(my_ips,j,k)
           endif
           c(my_ips-2,j,k) = c(my_ips-1,j,k)

        end do
     end do
       end if
       if (my_E_proc.eq.-1) then
               !$acc parallel loop gang vector collapse(2)  default(none)
         do k = my_kps,my_kpe
         do j = my_jps,my_jpe
#if defined WITH_ACC
           if (ADS_u(my_ibe,j,k).lt.0.0_rp) then
#else
           if (CB%u(my_ibe,j,k).lt.0.0_rp) then
#endif
             c(my_ipe+1,j,k) = 2.0_rp*CB%bvalue(2)-c(my_ipe,j,k)
           else
             c(my_ipe+1,j,k) = c(my_ipe,j,k)
           endif
           c(my_ipe+2,j,k) = c(my_ipe+1,j,k)

         end do
         end do
       end if
     end if
     ! In Y dimension
     if (mydim.eq.2) then
       if (my_S_proc.eq.-1) then
               !$acc parallel loop gang vector collapse(2)  default(none)
         do k = my_kps,my_kpe
         do i = my_ips,my_ipe
#if defined WITH_ACC
           if (ADS_v(my_jbs,i,k).gt.0.0_rp) then
#else
           if (CB%v(my_jbs,i,k).gt.0.0_rp) then
#endif
             c(i,my_jps-1,k) = 2.0_rp*CB%bvalue(1)-c(i,my_jps,k)
           else
             c(i,my_jps-1,k) = c(i,my_jps,k)
           endif
           c(i,my_jps-2,k) = c(i,my_jps-1,k)

         end do
         end do
       end if
       if (my_N_proc.eq.-1) then
               !$acc parallel loop gang vector collapse(2)  default(none)
         do k = my_kps,my_kpe
         do i = my_ips,my_ipe
#if defined WITH_ACC
           if (ADS_v(my_jbe,i,k).lt.0.0_rp) then
#else
           if (CB%v(my_jbe,i,k).lt.0.0_rp) then
#endif
             c(i,my_jpe+1,k) = 2.0_rp*CB%bvalue(2)-c(i,my_jpe,k)
           else
             c(i,my_jpe+1,k) = c(i,my_jpe,k)
           endif
           c(i,my_jpe+2,k) = c(i,my_jpe+1,k)

         end do
         end do
       end if
     end if
     ! In Z dimension
     if (mydim.eq.3) then
     !$acc data present(cb%iw)
       if (my_D_proc.eq.-1) then
               !$acc parallel loop gang vector collapse(2)  default(none)
         do j = my_jps,my_jpe
         do i = my_ips,my_ipe
#if defined WITH_ACC
           if (CB%iw(my_kbs,i,j).gt.0.0_rp) then
#else
           if (CB%w(my_kbs,i,j).gt.0.0_rp) then
#endif
             c(i,j,my_kps-1) = 2.0_rp*CB%bvalue(1)-c(i,j,my_kps)
           else
             c(i,j,my_kps-1) = c(i,j,my_kps)
           endif
           c(i,j,my_kps-2) = c(i,j,my_kps-1)

         end do
         end do
       end if
       if (my_U_proc.eq.-1) then
               !$acc parallel loop gang vector collapse(2)  default(none)
         do j = my_jps,my_jpe
         do i = my_ips,my_ipe
#if defined WITH_ACC
           if (CB%iw(my_kbe,i,j).lt.0.0_rp) then
#else
           if (CB%w(my_kbe,i,j).lt.0.0_rp) then
#endif
             c(i,j,my_kpe+1) = 2.0_rp*CB%bvalue(2)-c(i,j,my_kpe)
           else
             c(i,j,my_kpe+1) = c(i,j,my_kpe)
           endif
           c(i,j,my_kpe+2) = c(i,j,my_kpe+1)

         end do
         end do
       end if
     !$acc end data
     end if
     !$acc end data

     end subroutine freeflow
!
!
!
     subroutine dirichlet( c, mydim )
!******************************************************************
!*
!*   Apply Dirichlet boundary conditions
!*
!******************************************************************
     implicit none
     real   (rp), intent(inout) :: c  (my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h,my_kps_2h:my_kpe_2h)
     integer(ip), intent(in)    :: mydim

     integer(ip)                :: i,j,k
     ! In X dimension
     if (mydim.eq.1) then
       if (my_W_proc.eq.-1) then
         do k = my_kps,my_kpe
         do j = my_jps,my_jpe
           c(my_ips-1,j,k) = 2.0_rp*CB%bvalue(1)-c(my_ips,j,k)
           c(my_ips-2,j,k) = c(my_ips-1,j,k)
         end do
         end do
       end if
       if (my_E_proc.eq.-1) then
         do k = my_kps,my_kpe
         do j = my_jps,my_jpe
           c(my_ipe+1,j,k) = 2.0_rp*CB%bvalue(2)-c(my_ipe,j,k)
           c(my_ipe+2,j,k) = c(my_ipe+1,j,k)
         end do
         end do
       end if
     end if
     ! In Y dimension
     if (mydim.eq.2) then
       if (my_S_proc.eq.-1) then
         do k = my_kps,my_kpe
         do i = my_ips,my_ipe
           c(i,my_jps-1,k) = 2.0_rp*CB%bvalue(1)-c(i,my_jps,k)
           c(i,my_jps-2,k) = c(i,my_jps-1,k)
         end do
         end do
       end if
       if (my_N_proc.eq.-1) then
         do k = my_kps,my_kpe
         do i = my_ips,my_ipe
           c(i,my_jpe+1,k) = 2.0_rp*CB%bvalue(2)-c(i,my_jpe,k)
           c(i,my_jpe+2,k) = c(i,my_jpe+1,k)
         end do
         end do
       end if
     end if
     ! In Z dimension
     if (mydim.eq.3) then
       if (my_D_proc.eq.-1) then
         do j = my_jps,my_jpe
         do i = my_ips,my_ipe
           c(i,j,my_kps-1) = 2.0_rp*CB%bvalue(1)-c(i,j,my_kps)
           c(i,j,my_kps-2) = c(i,j,my_kps-1)
         end do
         end do
       end if
       if (my_U_proc.eq.-1) then
         do j = my_jps,my_jpe
         do i = my_ips,my_ipe
           c(i,j,my_kpe+1) = 2.0_rp*CB%bvalue(2)-c(i,j,my_kpe)
           c(i,j,my_kpe+2) = c(i,j,my_kpe+1)
         end do
         end do
       end if
     end if

     end subroutine dirichlet
!
!
!
     subroutine periodic( c, mydim  )
!******************************************************************
!*
!*   Apply Periodic boundary conditions, JUST FOR SERIAL
!*
!******************************************************************
     implicit none
     real   (rp), intent(inout) :: c  (my_ips_2h:my_ipe_2h,my_jps_2h:my_jpe_2h,my_kps_2h:my_kpe_2h)
     integer(ip), intent(in)    :: mydim

     integer(ip)               :: i,j,k
     ! In X dimension
     if (mydim.eq.1) then
       do k = my_kps,my_kpe
       do j = my_jps,my_jpe
         c(my_ips-1,j,k) = c(my_ips,j,k)
         c(my_ips-2,j,k) = c(my_ips-1,j,k)
         c(my_ipe+1,j,k) = c(my_ipe,j,k)
         c(my_ipe+2,j,k) = c(my_ipe+1,j,k)
       end do
       end do
     end if
     ! In Y dimension
     if (mydim.eq.2) then
       do k = my_kps,my_kpe
       do i = my_ips,my_ipe
         c(i,my_jps-1,k) = c(i,my_jps,k)
         c(i,my_jps-2,k) = c(i,my_jps-1,k)
         c(i,my_jpe+1,k) = c(i,my_jpe,k)
         c(i,my_jpe+2,k) = c(i,my_jpe+1,k)
       end do
       end do
     end if
     ! In Z dimension
     if (mydim.eq.3) then
       do j = my_jps,my_jpe
       do i = my_ips,my_ipe
         c(i,j,my_kps-1) = c(i,j,my_kps)
         c(i,j,my_kps-2) = c(i,j,my_kps-1)
         c(i,j,my_kpe+1) = c(i,j,my_kpe)
         c(i,j,my_kpe+2) = c(i,j,my_kpe+1)
       end do
       end do
     end if

     end subroutine periodic
!
!
!
     END MODULE ADS
